from django.contrib.auth import get_user_model
from django.contrib.auth.models import AnonymousUser
from django.test import SimpleTestCase, override_settings
from django.test.client import RequestFactory

from djangocms_oidc import utils


class TestUtils(SimpleTestCase):

    @override_settings(EXAMPLE_VARIABLE='example_value')
    def test_get_settings(self):
        value = utils.get_settings('EXAMPLE_VARIABLE')
        self.assertEqual(value, 'example_value')

    @override_settings(CMS_CACHE_PREFIX='prefix:')
    def test_get_cache_key(self):
        value = utils.get_cache_key("key")
        self.assertEqual(value, 'prefix:key')

    def test_only_user_is_staff_no_user(self):
        request = RequestFactory().request()
        context = {'request': request}
        self.assertFalse(utils.only_user_is_staff(context, None, None, None))

    def test_only_user_is_staff_user_is_not_staff(self):
        request = RequestFactory().request()
        request.user = get_user_model()()
        context = {'request': request}
        self.assertFalse(utils.only_user_is_staff(context, None, None, None))

    def test_only_user_is_staff(self):
        request = RequestFactory().request()
        request.user = get_user_model()(is_staff=True)
        context = {'request': request}
        self.assertTrue(utils.only_user_is_staff(context, None, None, None))

    def test_only_authenticated_user_no_user(self):
        request = RequestFactory().request()
        context = {'request': request}
        self.assertFalse(utils.only_authenticated_user(context, None, None, None))

    def test_only_authenticated_user_is_not_authenticated(self):
        request = RequestFactory().request()
        request.user = AnonymousUser()
        context = {'request': request}
        self.assertFalse(utils.only_authenticated_user(context, None, None, None))

    def test_only_authenticated_user(self):
        request = RequestFactory().request()
        request.user = get_user_model()()
        context = {'request': request}
        self.assertTrue(utils.only_authenticated_user(context, None, None, None))

    def test_email_verified_user_info_is_none(self):
        context = {}
        self.assertFalse(utils.email_verified(context, None, None, None))
        self.assertEqual(context, {})

    def test_email_verified(self):
        context = {}
        user_info = {
            'email': 'foo@foo.foo',
            'email_verified': True,
        }
        self.assertTrue(utils.email_verified(context, None, None, user_info))
        self.assertEqual(context, {})

    def test_email_verified_email_missing(self):
        context = {}
        user_info = {
            'email_verified': True,
        }
        self.assertFalse(utils.email_verified(context, None, None, user_info))
        self.assertEqual(context, {
            'dedicated_content': '''<ul class='messagelist'><li class='error'>Email missing.</li></ul>'''
        })

    def test_email_verified_email_verified_missing(self):
        context = {}
        user_info = {
            'email': 'foo@foo.foo',
        }
        self.assertFalse(utils.email_verified(context, None, None, user_info))
        self.assertEqual(context, {
            'dedicated_content': '''<ul class='messagelist'><li class='error'>Email is not verified.</li></ul>'''
        })

    def test_email_verified_all_missing(self):
        context = {}
        user_info = {}
        self.assertFalse(utils.email_verified(context, None, None, user_info))
        self.assertEqual(context, {
            'dedicated_content': '''<ul class='messagelist'><li class='error'>Email missing. '''
                                 '''Email is not verified.</li></ul>'''
        })



import json

from django.forms.widgets import Textarea


class JsonDataTextarea(Textarea):

    def format_value(self, value):
        """
        Return a value as it should appear when rendered in a template.
        """
        if value == "" or value is None:
            return None
        # Reformat json data with indentation.
        try:
            return json.dumps(json.loads(value), indent=4)
        except json.JSONDecodeError:
            pass
        return value



import logging
import time
from urllib.parse import urlencode

from django.contrib import messages
from django.http import HttpResponseRedirect, JsonResponse
from django.urls import reverse
from django.utils.crypto import get_random_string
from django.utils.translation import gettext_lazy as _
from mozilla_django_oidc.middleware import SessionRefresh
from mozilla_django_oidc.utils import absolutify

from .helpers import get_consumer, request_is_ajax

LOGGER = logging.getLogger("djangocms_oidc")


class OIDCSessionRefresh(SessionRefresh):

    def process_request(self, request):
        if not self.is_refreshable_url(request):
            LOGGER.debug('request is not refreshable')
            return

        expiration = request.session.get('oidc_id_token_expiration', 0)
        now = time.time()
        if expiration > now:
            # The id_token is still valid, so we don't have to do anything.
            LOGGER.debug('id token is still valid (%s > %s)', expiration, now)
            return

        LOGGER.debug('id token has expired')
        # The id_token has expired, so we have to re-authenticate silently.
        consumer = get_consumer(request)
        if consumer is None:
            LOGGER.debug('OIDC Consumer was not found in session.')
            messages.error(request, _('OIDC Consumer activation falied. Please try again.'))
            return HttpResponseRedirect(request.META.get('HTTP_REFERER', '/'))

        state = get_random_string(self.get_settings('OIDC_STATE_SIZE', 32))

        # Build the parameters as if we were doing a real auth handoff, except
        # we also include prompt=none.
        params = {
            'response_type': 'code',
            'client_id': consumer.provider.get_client_id(),
            'redirect_uri': absolutify(
                request,
                reverse(self.get_settings('OIDC_AUTHENTICATION_CALLBACK_URL',
                                          'oidc_authentication_callback'))
            ),
            'state': state,
            'scope': self.get_settings('OIDC_RP_SCOPES', 'openid email'),
            'prompt': 'none',
        }

        if self.get_settings('OIDC_USE_NONCE', True):
            nonce = get_random_string(self.get_settings('OIDC_NONCE_SIZE', 32))
            params.update({
                'nonce': nonce
            })
            request.session['oidc_nonce'] = nonce

        request.session['oidc_state'] = state
        if consumer.redirect_page is None:
            request.session['oidc_login_next'] = request.get_full_path()
        else:
            request.session['oidc_login_next'] = consumer.redirect_page.get_absolute_url()

        query = urlencode(params)
        redirect_url = f'{consumer.provider.authorization_endpoint}?{query}'
        if request_is_ajax(request):
            # Almost all XHR request handling in client-side code struggles
            # with redirects since redirecting to a page where the user
            # is supposed to do something is extremely unlikely to work
            # in an XHR request. Make a special response for these kinds
            # of requests.
            # The use of 403 Forbidden is to match the fact that this
            # middleware doesn't really want the user in if they don't
            # refresh their session.
            response = JsonResponse({'refresh_url': redirect_url}, status=403)
            response['refresh_url'] = redirect_url
            return response
        return HttpResponseRedirect(redirect_url)



import datetime
import logging

import requests
from cms.models.fields import PageField
from cms.models.pluginmodel import CMSPlugin
from django.conf import settings
from django.core.cache import cache
from django.core.exceptions import ValidationError
from django.db import models
from django.utils import timezone
from django.utils.translation import gettext_lazy as _
from multiselectfield import MultiSelectField
from requests.exceptions import RequestException

from .utils import DEFAULT_DISPLAY_CONTENT, get_cache_key, get_settings

HANDOVER_DATA = 'handover'
LOGIN_USER = 'login'

logger = logging.getLogger("djangocms_oidc")


class ConsumerRegistrationExpired(Exception):
    """Consumer registration expired."""


class OIDCRegisterConsumer(CMSPlugin):

    name = models.CharField(
        verbose_name=_('Name'), max_length=255, unique=True, default='Default',
        help_text=_("Registration name"))
    register_url = models.URLField(
        verbose_name=_('Register consumer URL'),
        help_text=_("Register consumer automaticaly, when it is not. (Client ID and secret missing)."))
    # TODO: More parameters:
    # application_type" "web"
    # redirect_uris: (given from mozilla-oidc urls)
    # request_object_signing_alg: OIDC_RP_SIGN_ALGO (HS256)
    # token_endpoint_auth_method: "client_secret_post"
    # response_types: ["code"]
    # client_name: (same as provider name)
    # logo_uri:

    def __str__(self):
        return self.name

    def get_payload(self, name, redirect_uris):
        return {
            "application_type": "web",
            "id_token_signed_response_alg": get_settings('OIDC_RP_SIGN_ALGO', 'HS256'),
            "redirect_uris": redirect_uris,
            "token_endpoint_auth_method": "client_secret_post",
            "response_types": ["code"],
            "client_name": name,
        }

    def make_registration(self, name, redirect_uris):
        response = requests.post(self.register_url, data=self.get_payload(name, redirect_uris))
        response.raise_for_status()
        return response.json()
        # return response.json(): {
        #     'client_id_issued_at': 1598961156,
        #     'response_types': ['code'],
        #     'request_object_signing_alg': 'HS256',
        #     'registration_client_uri': 'https://mojeid.regtest.nic.cz/oidc/registration/?client_id=1d9G0Oid4E5V',
        #     'application_type': 'web',
        #     'registration_access_token': 'zhbvda8JZgsamO8Lm0TgZ0UA50cLXrZm',
        #     'redirect_uris': ['http://localhost:8180/oidc/callback/'],
        #     'client_secret_expires_at': 1599047556,
        #     'client_secret': '5ed3a93b14cb5f1fdb97cb3dde1daddade443753eeb84432320b78a2',
        #     'client_name': 'MojeID Test - Dynamic consumer',
        #     'client_id': '1d9G0Oid4E5V'
        # }


class OIDCProvider(CMSPlugin):

    name = models.CharField(verbose_name=_('Name'), max_length=255, unique=True, help_text=_("Provider name"))
    slug = models.SlugField(verbose_name=_('Code'),
                            help_text=_("Used, for example, as a css class for a button or link."))
    client_id = models.CharField(
        verbose_name=_('Client ID'), max_length=255, null=True, blank=True,
        help_text=_('The value is required if "Register consumer" is not set.'))
    client_secret = models.CharField(
        verbose_name=_('Client secret'), max_length=255, null=True, blank=True,
        help_text=_('The value is required if "Register consumer" is not set.'))
    register_consumer = models.ForeignKey(
        OIDCRegisterConsumer, on_delete=models.CASCADE, null=True, blank=True, verbose_name=_('Register consumer'),
        help_text=_("Register consumer on the provider, when Client ID and secret are not set."))
    authorization_endpoint = models.URLField(verbose_name=_('Authorization endpoint'))
    token_endpoint = models.URLField(verbose_name=_('Token endpoint'))
    user_endpoint = models.URLField(verbose_name=_('User endpoint'))
    account_url = models.URLField(verbose_name=_('Account URL'), null=True, blank=True)
    logout_url = models.URLField(verbose_name=_('Logout URL'), null=True, blank=True)

    progress_code = "--PROGRESS--"

    def __str__(self):
        return self.name

    def _get_cached(self, key):
        data = cache.get(self.get_cache_key())
        if data is None:
            raise ConsumerRegistrationExpired()
        return data[key]

    def get_client_id(self):
        if self.client_id:
            return self.client_id
        return self._get_cached('client_id')

    def get_client_secret(self):
        if self.client_secret:
            return self.client_secret
        return self._get_cached('client_secret')

    def needs_register(self):
        if self.client_id:
            return False
        return cache.get(self.get_cache_key()) is None

    def get_cache_key(self):
        return get_cache_key(f"djangocms_oidc_provider:{self.pk}")

    def registration_in_progress(self):
        return cache.get(self.get_cache_key()) == self.progress_code

    def register_consumer_into_cache(self, redirect_uris):
        logger.info(f"Register consumer at {self.register_consumer.register_url}")
        key = self.get_cache_key()
        cache.set(key, self.progress_code, 10)  # Assign the key for 10 seconds.
        try:
            data = self.register_consumer.make_registration(self.name, redirect_uris)
        except RequestException as msg:
            logger.error(msg)
            return msg
        logger.info(data)
        if data['client_secret_expires_at'] == 0:
            expires_at = "never"
            duration = None
        else:
            expires_at = datetime.datetime.fromtimestamp(data['client_secret_expires_at'])
            expires_at = timezone.make_aware(expires_at, timezone.utc)
            duration = expires_at.timestamp() - timezone.now().timestamp()
        logger.info(f"Expires at {expires_at}")
        content = {
            'client_id': data['client_id'],
            'client_secret': data['client_secret'],
            'expires_at': expires_at,
        }
        cache.set(key, content, duration)
        logger.debug(f"Store into cache: {key} for {duration} sec.")
        return None

    def get_registration_consumer_info(self):
        data = {
            'client_id': None,
            'expires_at': None,
        }
        if self.client_id:
            data['consumer_type'] = 'MANAGED'
            data['client_id'] = self.client_id
        else:
            data['consumer_type'] = 'AUTOMATIC'
            info = cache.get(self.get_cache_key())
            if info is not None:
                data['client_id'] = info.get('client_id')
                data['expires_at'] = info.get('expires_at')
        return data


def validate_claims(value):
    """Validate Claims."""
    if not isinstance(value, dict):
        raise ValidationError(_("The value must be a dictionary type: {}"))
    if "userinfo" in value:
        if not isinstance(value["userinfo"], dict):
            raise ValidationError(_('The value "userinfo" must be a dictionary type: {"userinfo": ...}'))
        for key, datadict in value["userinfo"].items():
            if not isinstance(datadict, dict):
                raise ValidationError(
                    _('The "%(key)s" must be a dictionary type: "%(key)s": {"essential": true},'),
                    params={'key': key}
                )
            if "essential" not in datadict:
                raise ValidationError(
                    _('The "%(key)s" value must contain key "essential": "%(key)s": {"essential": true},'),
                    params={'key': key}
                )
            if not isinstance(datadict["essential"], bool):
                raise ValidationError(
                    _('The value of "essential" must be a boolean type": "%(key)s": {"essential": true},'),
                    params={'key': key}
                )


class OIDCHandoverDataBase(CMSPlugin):

    consumer_type = HANDOVER_DATA
    AUTHORIZATION_PROMPT = (
        ('none', _('No interaction')),
        ('login', _('Force login')),
        ('consent', _('Force consent with handovered data')),
    )

    provider = models.ForeignKey(OIDCProvider, on_delete=models.CASCADE)
    button_label = models.CharField(
        verbose_name=_('Button label'), max_length=80, null=True, blank=True,
        help_text=_("Button text for unlogged in user."))
    claims = models.JSONField(
        verbose_name=_("Claims"), validators=[validate_claims], help_text=_("Claims attributes for data handover."),
        default={"userinfo": {}})
    insist_on_required_claims = models.BooleanField(
        verbose_name=_("Insist on required claims"), default=False,
        help_text=_("Consider the data invalid if not all the required data has been handovered."))
    verified_by = models.CharField(
        verbose_name=_('Verified by names'), max_length=255, null=True, blank=True,
        help_text=_("Verified by names (separated by space)."))
    redirect_page = PageField(verbose_name=_('CMS Page'), blank=True, null=True, on_delete=models.SET_NULL,
                              help_text=_("Redirect to the page after login or hangover data."))
    authorization_prompt = MultiSelectField(
        verbose_name=_('Prompt'), max_length=255, null=True, blank=True, choices=AUTHORIZATION_PROMPT,
        help_text=_('Prompt for user at authorization. "No interaction" cannot be combined with others.'))

    class Meta:
        abstract = True

    def __str__(self):
        if self.button_label:
            name = self.button_label
        else:
            try:
                name = self.provider.name
            except OIDCProvider.DoesNotExist:
                name = '[No provider yet]'
        return name

    def can_login(self):
        return self.consumer_type == LOGIN_USER


class OIDCLoginBase(OIDCHandoverDataBase):
    """OIDC Consumer for login user."""

    consumer_type = LOGIN_USER

    class Meta:
        abstract = True

    no_new_user = models.BooleanField(
        verbose_name=_("No new user"), default=False,
        help_text=_("Do not create a new user account if the user is not recognized."))


class OIDCHandoverData(OIDCHandoverDataBase):
    """OIDC Consumer for data handover."""


class OIDCLogin(OIDCLoginBase):
    """OIDC Consumer for login user."""


CONSUMER_CLASS = {
    HANDOVER_DATA: OIDCHandoverData,
    LOGIN_USER: OIDCLogin,
}


class OIDCIdentifier(CMSPlugin):
    user = models.OneToOneField(getattr(settings, 'AUTH_USER_MODEL', 'auth.User'), on_delete=models.CASCADE)
    provider = models.ForeignKey(OIDCProvider, on_delete=models.CASCADE)
    uident = models.CharField(verbose_name=_('Identifier'), max_length=255, unique=True,
                              help_text=_("Provider unique identifier."))


def get_display_content_settings():
    """Get a list of functions that are used for check conditions for displaying dedicated content."""
    return getattr(settings, 'DJANGOCMS_OIDC_DISPLAY_CONTENT', DEFAULT_DISPLAY_CONTENT)


class OIDCDisplayDedicatedContent(CMSPlugin):
    conditions = models.CharField(
        verbose_name=_('Conditions'),
        help_text=_('Show content only if the given conditions are met.'),
        max_length=255, null=True, blank=True,
        choices=[(code, label) for code, label, _ in get_display_content_settings()],
    )


class OIDCShowAttribute(CMSPlugin):
    """Show attribute."""
    verified_by = models.CharField(
        verbose_name=_('Verified by names'), max_length=255, default="name",
        help_text=_("Verified by names (separated by space). Eg.: "
                    "name given_name+family_name preferred_username nickname email openid2_id"))
    default_value = models.CharField(
        verbose_name=_('Default value'), max_length=255, null=True, blank=True,
        help_text=_("Show this text when attribute has not been handovered."))
    css_names = models.CharField(
        verbose_name=_('CSS names'), max_length=255, null=True, blank=True,
        help_text=_("Extra class names (separated by space)."))

    def __str__(self):
        return self.verified_by



from collections import OrderedDict
from unittest.mock import patch
from urllib.parse import urlencode

import requests_mock
from cms.api import create_page
from cms.test_utils.testcases import CMSTestCase
from django.contrib import messages
from django.contrib.auth import get_user_model
from django.contrib.messages import get_messages
from django.contrib.messages.storage.fallback import FallbackStorage
from django.contrib.sessions.backends.signed_cookies import SessionStore
from django.core import cache
from django.http import JsonResponse
from django.test import RequestFactory, override_settings
from django.urls import reverse
from freezegun import freeze_time

from djangocms_oidc.constants import DJANGOCMS_PLUGIN_SESSION_KEY, DJANGOCMS_USER_SESSION_KEY
from djangocms_oidc.models import OIDCHandoverData, OIDCIdentifier, OIDCLogin, OIDCProvider, OIDCRegisterConsumer
from djangocms_oidc.views import DjangocmsOIDCAuthenticationCallbackView, DjangocmsOIDCAuthenticationRequestView


class CollectMessagesMixin:

    def _get_messages(self, request):
        return [(msg.level, msg.message) for msg in get_messages(request)]


@override_settings(CMS_CACHE_PREFIX='prefix:')
@override_settings(CACHES={'default': {'BACKEND': 'django.core.cache.backends.locmem.LocMemCache'}})
class TestOIDCSignupView(CollectMessagesMixin, CMSTestCase):

    consumer_registration = {
        'client_id': '1d9G0Oid4E5V',
        'client_secret': '5ed3a93b14cb5f1fdb97cb3dde1daddade443753eeb84432320b78a2',
        'expires_at': 'never'
    }

    @classmethod
    def setUpTestData(cls):
        cls.register_consumer = OIDCRegisterConsumer.objects.create(
            name="Test", register_url="https://foo.foo/register")
        cls.provider = OIDCProvider.objects.create(
            name="Provider", slug="provider", register_consumer=cls.register_consumer,
            authorization_endpoint="https://foo.foo/authorization-endpoint")
        cls.consumer = OIDCHandoverData.objects.create(provider=cls.provider, claims={})

    def setUp(self):
        cache.cache.clear()

    def test_signup_activation_failed(self):
        url = reverse('djangocms_oidc_signup', kwargs={'consumer_type': 'handover', 'plugin_id': 42})
        response = self.client.get(url)
        self.assertRedirects(response, '/')
        self.assertEqual(self._get_messages(response.wsgi_request), [
            (messages.ERROR, 'OIDC Consumer activation falied. Please try again.')
        ])
        data = cache.cache.get(f'prefix:djangocms_oidc_provider:{self.provider.pk}')
        self.assertIsNone(data)

    def test_signup_progress(self):
        cache.cache.set(f'prefix:djangocms_oidc_provider:{self.provider.pk}', '--PROGRESS--')
        url = reverse('djangocms_oidc_signup', kwargs={
            'consumer_type': self.consumer.consumer_type, 'plugin_id': self.consumer.pk})
        response = self.client.get(url)
        self.assertRedirects(response, '/')
        self.assertEqual(self._get_messages(response.wsgi_request), [
            (messages.INFO, 'Communication with the provider is in progress. Please try again later.')
        ])
        data = cache.cache.get(f'prefix:djangocms_oidc_provider:{self.provider.pk}')
        self.assertEqual(data, '--PROGRESS--')

    @requests_mock.Mocker()
    def test_provider_needs_register(self, mock_req):
        mock_req.post("https://foo.foo/register", text='Not Found', status_code=404)
        url = reverse('djangocms_oidc_signup', kwargs={
            'consumer_type': self.consumer.consumer_type, 'plugin_id': self.consumer.pk})
        response = self.client.get(url)
        self.assertRedirects(response, '/')
        self.assertEqual(self._get_messages(response.wsgi_request), [
            (messages.ERROR, 'Communication with the provider failed. Please try again later.'),
            (messages.ERROR, '404 Client Error: None for url: https://foo.foo/register'),
        ])
        data = cache.cache.get(f'prefix:djangocms_oidc_provider:{self.provider.pk}')
        self.assertEqual(data, '--PROGRESS--')

    @requests_mock.Mocker()
    def test_provider_make_registation(self, mock_req):
        mock_req.post("https://foo.foo/register", json={
            'client_secret_expires_at': 0,
            'client_secret': '5ed3a93b14cb5f1fdb97cb3dde1daddade443753eeb84432320b78a2',
            'client_id': '1d9G0Oid4E5V'
        })
        url = reverse('djangocms_oidc_signup', kwargs={
            'consumer_type': self.consumer.consumer_type, 'plugin_id': self.consumer.pk, 'prompt': 'login'})
        response = self.client.get(url)
        self.assertRedirects(response, "{}?next=/&prompt=login".format(reverse("oidc_authentication_init")),
                             target_status_code=302)
        self.assertEqual(self._get_messages(response.wsgi_request), [])
        data = cache.cache.get(f'prefix:djangocms_oidc_provider:{self.provider.pk}')
        self.assertEqual(data, self.consumer_registration)

    def test_provider_already_registered(self):
        cache.cache.set(f'prefix:djangocms_oidc_provider:{self.provider.pk}', self.consumer_registration)
        url = reverse('djangocms_oidc_signup', kwargs={
            'consumer_type': self.consumer.consumer_type, 'plugin_id': self.consumer.pk})
        session = self.client.session
        session[DJANGOCMS_USER_SESSION_KEY] = 42
        session.save()
        response = self.client.get(url)
        self.assertIsNone(self.client.session.get(DJANGOCMS_USER_SESSION_KEY))
        self.assertRedirects(response, "{}?next=/".format(reverse("oidc_authentication_init")), target_status_code=302)
        self.assertEqual(self._get_messages(response.wsgi_request), [])
        data = cache.cache.get(f'prefix:djangocms_oidc_provider:{self.provider.pk}')
        self.assertEqual(data, self.consumer_registration)


class TestOIDCDismissView(CMSTestCase):

    def test_dismiss(self):
        url = reverse('djangocms_oidc_dismiss')
        response = self.client.get(url)
        self.assertRedirects(response, '/')

    def test_dismiss_plugin_key(self):
        url = reverse('djangocms_oidc_dismiss')
        session = self.client.session
        session[DJANGOCMS_PLUGIN_SESSION_KEY] = 42
        session.save()
        response = self.client.get(url)
        self.assertRedirects(response, '/')
        self.assertIsNone(self.client.session.get(DJANGOCMS_PLUGIN_SESSION_KEY))

    def test_dismiss_user_key(self):
        url = reverse('djangocms_oidc_dismiss')
        session = self.client.session
        session[DJANGOCMS_USER_SESSION_KEY] = 42
        session.save()
        response = self.client.get(url)
        self.assertRedirects(response, '/')
        self.assertIsNone(self.client.session.get(DJANGOCMS_USER_SESSION_KEY))


class TestOIDCLogoutView(CMSTestCase):

    def test_logout(self):
        url = reverse('djangocms_oidc_logout')
        response = self.client.get(url)
        self.assertRedirects(response, '/')

    def test_logout_authenticated_user(self):
        user = get_user_model().objects.create(username="user")
        self.client.force_login(user)
        url = reverse('djangocms_oidc_logout')
        response = self.client.get(url)
        self.assertRedirects(response, '/')
        self.assertFalse(response.wsgi_request.user.is_authenticated)


class TestOIDCDeleteIdentifiersView(CollectMessagesMixin, CMSTestCase):

    @classmethod
    def setUpTestData(cls):
        cls.user = get_user_model().objects.create(username="user")
        cls.provider = OIDCProvider.objects.create(name="Provider", slug="provider")

    def test_empty_post(self):
        url = reverse('djangocms_oidc_delete_identifiers')
        response = self.client.post(url)
        self.assertRedirects(response, '/')
        self.assertEqual(self._get_messages(response.wsgi_request), [])

    def test_not_authenticated_user(self):
        url = reverse('djangocms_oidc_delete_identifiers')
        response = self.client.post(url, {"djangocms_oidc_delete_identifier": "yes"})
        self.assertRedirects(response, '/')
        self.assertEqual(self._get_messages(response.wsgi_request), [])

    def test_no_identifier_selected(self):
        self.client.force_login(self.user)
        url = reverse('djangocms_oidc_delete_identifiers')
        response = self.client.post(url, {
            "djangocms_oidc_delete_identifier": "yes",
            "form-TOTAL_FORMS": "0",
            "form-INITIAL_FORMS": "0",
            "form-MAX_NUM_FORMS": "1000",
        })
        self.assertRedirects(response, '/')
        self.assertEqual(self._get_messages(response.wsgi_request), [
            (messages.INFO, 'No identifier has been deleted. Check one to delete.'),
        ])

    def test_invlaid_post(self):
        self.client.force_login(self.user)
        url = reverse('djangocms_oidc_delete_identifiers')
        response = self.client.post(url, {
            "djangocms_oidc_delete_identifier": "yes",
            "form-TOTAL_FORMS": "1",
            "form-INITIAL_FORMS": "1",
            "form-MAX_NUM_FORMS": "1000",
        })
        self.assertRedirects(response, '/')
        self.assertEqual(self._get_messages(response.wsgi_request), [
            (messages.ERROR, '* cmsplugin_ptr\n  * This field is required.'),
        ])

    def test_invlaid_post_ajax(self):
        self.client.force_login(self.user)
        url = reverse('djangocms_oidc_delete_identifiers')
        response = self.client.post(url, {
            "djangocms_oidc_delete_identifier": "yes",
            "form-TOTAL_FORMS": "1",
            "form-INITIAL_FORMS": "1",
            "form-MAX_NUM_FORMS": "1000",
        }, HTTP_X_REQUESTED_WITH='XMLHttpRequest')
        self.assertIsInstance(response, JsonResponse)
        self.assertEqual(response.json(), {
            'result': 'ERROR',
            'messages': [{'cmsplugin_ptr': ['This field is required.']}],
        })
        self.assertEqual(self._get_messages(response.wsgi_request), [])

    def test_delete_identifier(self):
        other_user = get_user_model().objects.create(username="other")
        ident = OIDCIdentifier.objects.create(user=self.user, provider=self.provider, uident="1234567890")
        OIDCIdentifier.objects.create(user=other_user, provider=self.provider, uident="9876543210")
        self.client.force_login(self.user)
        url = reverse('djangocms_oidc_delete_identifiers')
        response = self.client.post(url, {
            "djangocms_oidc_delete_identifier": "yes",
            "form-0-cmsplugin_ptr": [ident.cmsplugin_ptr_id],
            "form-0-DELETE": "on",
            "form-TOTAL_FORMS": "1",
            "form-INITIAL_FORMS": "1",
            "form-MAX_NUM_FORMS": "1000",
        })
        self.assertRedirects(response, '/')
        self.assertEqual(self._get_messages(response.wsgi_request), [
            (messages.SUCCESS, 'Identifier has been deleted.'),
        ])
        self.assertQuerysetEqual(OIDCIdentifier.objects.all().values_list('uident', flat=True), [
            '9876543210'], transform=str)

    def test_delete_identifier_by_ajax(self):
        other_user = get_user_model().objects.create(username="other")
        ident = OIDCIdentifier.objects.create(user=self.user, provider=self.provider, uident="1234567890")
        OIDCIdentifier.objects.create(user=other_user, provider=self.provider, uident="9876543210")
        self.client.force_login(self.user)
        url = reverse('djangocms_oidc_delete_identifiers')
        response = self.client.post(url, {
            "djangocms_oidc_delete_identifier": "yes",
            "form-0-cmsplugin_ptr": [ident.cmsplugin_ptr_id],
            "form-0-DELETE": "on",
            "form-TOTAL_FORMS": "1",
            "form-INITIAL_FORMS": "1",
            "form-MAX_NUM_FORMS": "1000",
        }, HTTP_X_REQUESTED_WITH='XMLHttpRequest')
        self.assertIsInstance(response, JsonResponse)
        self.assertEqual(response.json(), {'result': 'SUCCESS', 'messages': ['Identifier has been deleted.']})
        self.assertEqual(self._get_messages(response.wsgi_request), [])


class CreateRequestMixin:

    def _create_request(self):
        request = RequestFactory().request()
        request.session = SessionStore()
        request._messages = FallbackStorage(request)
        return request


@override_settings(CMS_CACHE_PREFIX='prefix:')
@override_settings(CACHES={'default': {'BACKEND': 'django.core.cache.backends.locmem.LocMemCache'}})
class TestDjangocmsOIDCAuthenticationRequestView(CreateRequestMixin, CollectMessagesMixin, CMSTestCase):

    @classmethod
    def setUpTestData(cls):
        cls.provider = OIDCProvider.objects.create(
            name="Provider", slug="provider", client_id="1234567890", client_secret="secret",
            authorization_endpoint="https://foo.foo/authorization-endpoint")
        cls.consumer = OIDCHandoverData.objects.create(provider=cls.provider, claims={})

    def setUp(self):
        cache.cache.clear()

    def _create_request_get(self, params):
        request = RequestFactory().get(params)
        request.session = SessionStore()
        request._messages = FallbackStorage(request)
        return request

    def test_consumer_activation_falied(self):
        request = self._create_request()
        view = DjangocmsOIDCAuthenticationRequestView.as_view()
        response = view(request)
        self.assertRedirects(response, '/', fetch_redirect_response=False)
        self.assertEqual(self._get_messages(request), [
            (messages.ERROR, 'OIDC Consumer activation falied. Please try again.'),
        ])

    @freeze_time('2020-10-08 17:02:42')
    @patch('djangocms_oidc.views.get_random_string')
    def test_consumer(self, mock_views_random):
        mock_views_random.return_value = 'random'
        request = self._create_request()
        request.session[DJANGOCMS_PLUGIN_SESSION_KEY] = ('handover', self.consumer.pk)
        view = DjangocmsOIDCAuthenticationRequestView.as_view()
        response = view(request)
        self.assertEqual(self._get_messages(request), [])
        self.assertEqual(request.session.get('oidc_nonce'), 'random')
        self.assertIsNone(request.session.get('oidc_login_next'))
        self.assertEqual(request.session.get('oidc_states'), {'random': {'nonce': 'random', 'added_on': 1602176562.0}})
        query = OrderedDict([
            ('response_type', 'code'),
            ('scope', 'openid email'),
            ('client_id', '1234567890'),
            ('redirect_uri', 'http://testserver/callback/'),
            ('state', 'random'),
            ('nonce', 'random'),
            ('claims', '{}'),
        ])
        url = f"https://foo.foo/authorization-endpoint?{urlencode(query)}"
        self.assertRedirects(response, url, fetch_redirect_response=False)

    @override_settings(OIDC_USE_NONCE=False)
    @freeze_time('2020-10-08 17:02:42')
    @patch('djangocms_oidc.views.get_random_string')
    def test_consumer_nonce_none_and_redirect_page(self, mock_views_random):
        mock_views_random.return_value = 'random'
        cms_page = create_page('test', 'test_content_plugin.html', 'en', slug="test")
        self.consumer.redirect_page = cms_page
        self.consumer.save()
        request = self._create_request()
        request.session[DJANGOCMS_PLUGIN_SESSION_KEY] = ('handover', self.consumer.pk)
        view = DjangocmsOIDCAuthenticationRequestView.as_view()
        response = view(request)
        self.assertEqual(self._get_messages(request), [])
        self.assertIsNone(request.session.get('oidc_nonce'))
        self.assertEqual(request.session.get('oidc_login_next'), '/test/')
        self.assertEqual(request.session.get('oidc_states'), {'random': {'nonce': None, 'added_on': 1602176562.0}})
        query = OrderedDict([
            ('response_type', 'code'),
            ('scope', 'openid email'),
            ('client_id', '1234567890'),
            ('redirect_uri', 'http://testserver/callback/'),
            ('state', 'random'),
            ('claims', '{}'),
        ])
        url = f"https://foo.foo/authorization-endpoint?{urlencode(query)}"
        self.assertRedirects(response, url, fetch_redirect_response=False)

    def test_get_extra_params_consumer_is_none(self):
        request = self._create_request()
        authreq = DjangocmsOIDCAuthenticationRequestView()
        params = authreq.get_extra_params(request, None)
        self.assertEqual(params, {})

    @override_settings(OIDC_AUTH_REQUEST_EXTRA_PARAMS={'audience': 'some-api.example.com'})
    def test_get_extra_params_consumer_is_none_extra_params(self):
        request = self._create_request()
        authreq = DjangocmsOIDCAuthenticationRequestView()
        params = authreq.get_extra_params(request, None)
        self.assertEqual(params, {'audience': 'some-api.example.com'})

    def test_get_extra_params_consumer_with_claims(self):
        request = self._create_request()
        authreq = DjangocmsOIDCAuthenticationRequestView()
        consumer = OIDCHandoverData.objects.create(provider=self.provider, claims={})
        consumer.claims = {'email': {'essential': True}}
        params = authreq.get_extra_params(request, consumer)
        self.assertEqual(params, {'claims': '{"email": {"essential": true}}'})

    def test_get_extra_params_authorization_prompt(self):
        request = self._create_request()
        authreq = DjangocmsOIDCAuthenticationRequestView()
        consumer = OIDCHandoverData.objects.create(provider=self.provider, claims={})
        consumer.claims = {'email': {'essential': True}}
        consumer.authorization_prompt = ['login']
        params = authreq.get_extra_params(request, consumer)
        self.assertEqual(params, {
            'claims': '{"email": {"essential": true}}',
            'prompt': 'login',
        })

    def test_get_extra_params_authorization_prompt_get(self):
        request = self._create_request_get('?prompt=consent,login,foo,none')
        authreq = DjangocmsOIDCAuthenticationRequestView()
        consumer = OIDCHandoverData.objects.create(provider=self.provider, claims={})
        consumer.claims = {'email': {'essential': True}}
        consumer.authorization_prompt = ['login']
        params = authreq.get_extra_params(request, consumer)
        self.assertEqual(sorted(params.keys()), sorted(['claims', 'prompt']))
        self.assertEqual(params['claims'], '{"email": {"essential": true}}')
        self.assertEqual(sorted(params['prompt'].split(' ')), ['consent', 'login'])

    def test_get_extra_params_authorization_prompt_get_none(self):
        request = self._create_request_get('?prompt=none')
        authreq = DjangocmsOIDCAuthenticationRequestView()
        consumer = OIDCHandoverData.objects.create(provider=self.provider, claims={})
        consumer.claims = {'email': {'essential': True}}
        params = authreq.get_extra_params(request, consumer)
        self.assertEqual(params, {
            'claims': '{"email": {"essential": true}}',
            'prompt': 'none',
        })


class TestDjangocmsOIDCAuthenticationCallbackView(CreateRequestMixin, CollectMessagesMixin, CMSTestCase):

    @classmethod
    def setUpTestData(cls):
        cls.provider = OIDCProvider.objects.create(
            name="Provider", slug="provider", client_id="1234567890", client_secret="secret",
            authorization_endpoint="https://foo.foo/authorization-endpoint")
        cls.consumer = OIDCLogin.objects.create(provider=cls.provider, claims={})

    def test_failure_url(self):
        callback = DjangocmsOIDCAuthenticationCallbackView()
        callback.request = self._create_request()
        self.assertEqual(callback.failure_url, '/')
        self.assertEqual(self._get_messages(callback.request), [])

    def test_failure_url_message(self):
        callback = DjangocmsOIDCAuthenticationCallbackView()
        callback.request = self._create_request()
        callback.request.session[DJANGOCMS_PLUGIN_SESSION_KEY] = ('login', self.consumer.pk)
        self.assertEqual(callback.failure_url, '/')
        self.assertEqual(self._get_messages(callback.request), [(messages.INFO, 'Login failed.')])


class TestTestingView(CMSTestCase):

    def test_homepage(self):
        response = self.client.get(reverse("test_home_page"))
        self.assertContains(response, 'Test Home page.')

    def test_login(self):
        response = self.client.get(reverse("login"))
        self.assertContains(response, 'Test Login page.')

    def test_fake_view(self):
        response = self.client.get(reverse("mdo_fake_view"))
        self.assertContains(response, 'Win!')



# Generated by Django 2.2.13 on 2021-03-26 11:38

from django.db import migrations, models
import multiselectfield.db.fields


class Migration(migrations.Migration):

    dependencies = [
        ('djangocms_oidc', '0001_initial'),
    ]

    operations = [
        migrations.AddField(
            model_name='oidchandoverdata',
            name='button_label',
            field=models.CharField(blank=True, help_text='Button text for unlogged in user.', max_length=80, null=True, verbose_name='Button label'),
        ),
        migrations.AddField(
            model_name='oidclogin',
            name='button_label',
            field=models.CharField(blank=True, help_text='Button text for unlogged in user.', max_length=80, null=True, verbose_name='Button label'),
        ),
        migrations.AlterField(
            model_name='oidcdisplaydedicatedcontent',
            name='conditions',
            field=models.CharField(blank=True, choices=[('only_authenticated_user', 'Only authenticated user'), ('only_user_is_staff', 'Only user is staff'), ('email_verified', 'Email verified'), ('validated_account_to_submit_order', 'Validated account to submit order')], help_text='Show content only if the given conditions are met.', max_length=255, null=True, verbose_name='Conditions'),
        ),
        migrations.AlterField(
            model_name='oidchandoverdata',
            name='authorization_prompt',
            field=multiselectfield.db.fields.MultiSelectField(blank=True, choices=[('none', 'No interaction'), ('login', 'Force login'), ('consent', 'Force consent with handovered data')], help_text='Prompt for user at authorization. "No interaction" cannot be combined with others.', max_length=255, null=True, verbose_name='Prompt'),
        ),
        migrations.AlterField(
            model_name='oidclogin',
            name='authorization_prompt',
            field=multiselectfield.db.fields.MultiSelectField(blank=True, choices=[('none', 'No interaction'), ('login', 'Force login'), ('consent', 'Force consent with handovered data')], help_text='Prompt for user at authorization. "No interaction" cannot be combined with others.', max_length=255, null=True, verbose_name='Prompt'),
        ),
    ]



from django.contrib.admin.sites import site
from django.contrib.auth import get_user_model
from django.test import TestCase

from djangocms_oidc.admin.options import OIDCIdentifierAdmin
from djangocms_oidc.models import OIDCIdentifier


class TestAdminOptions(TestCase):

    def test_user_email(self):
        admin = OIDCIdentifierAdmin(admin_site=site, model=OIDCIdentifier)
        user = get_user_model()(email="email@foo.foo")
        obj = OIDCIdentifier(user=user)
        self.assertEqual(admin.user_email(obj), "email@foo.foo")



from django.urls import path, re_path
from django.utils.module_loading import import_string
from mozilla_django_oidc.utils import import_from_settings

from .models import CONSUMER_CLASS
from .views import OIDCDeleteIdentifiersView, OIDCLogoutView

DEFAULT_SIGNUP_VIEW_CLASS = 'djangocms_oidc.views.OIDCSignupView'
SIGNUP_VIEW_CLASS_PATH = import_from_settings('DJANGOCMS_OIDC_SIGNUP_VIEW_CLASS', DEFAULT_SIGNUP_VIEW_CLASS)
OIDCCSignupViewClass = import_string(SIGNUP_VIEW_CLASS_PATH)

DEFAULT_DISMISS_VIEW_CLASS = 'djangocms_oidc.views.OIDCDismissView'
DISMISS_VIEW_CLASS_PATH = import_from_settings('DJANGOCMS_OIDC_DISMISS_VIEW_CLASS', DEFAULT_DISMISS_VIEW_CLASS)
OIDCDismissViewClass = import_string(DISMISS_VIEW_CLASS_PATH)

urlpatterns = [
    re_path(r'^oidc-sign-up/(?P<consumer_type>({}))-(?P<plugin_id>\d+)(?P<prompt>[\w,]+)?/$'.format(
        '|'.join(CONSUMER_CLASS.keys())),
        OIDCCSignupViewClass.as_view(), name='djangocms_oidc_signup'),
    path('oidc-dismiss/', OIDCDismissViewClass.as_view(), name='djangocms_oidc_dismiss'),
    path('oidc-logout/', OIDCLogoutView.as_view(), name='djangocms_oidc_logout'),
    path('oidc-delete-identifiers/', OIDCDeleteIdentifiersView.as_view(), name='djangocms_oidc_delete_identifiers'),
]



import json
from unittest.mock import patch

import requests_mock
from django.contrib import messages
from django.contrib.auth import get_user_model
from django.contrib.auth.models import AnonymousUser
from django.contrib.messages import get_messages
from django.contrib.messages.storage.fallback import FallbackStorage
from django.core.exceptions import ImproperlyConfigured, SuspiciousOperation
from django.test import RequestFactory, TestCase, override_settings
from freezegun import freeze_time
from requests.exceptions import HTTPError

from djangocms_oidc.auth import DjangocmsOIDCAuthenticationBackend
from djangocms_oidc.constants import DJANGOCMS_PLUGIN_SESSION_KEY, DJANGOCMS_USER_SESSION_KEY
from djangocms_oidc.models import OIDCHandoverData, OIDCIdentifier, OIDCLogin, OIDCProvider


class TestAuthBackend(TestCase):

    @override_settings(OIDC_RP_SIGN_ALGO='RS256')
    def test_invalid_config(self):
        with self.assertRaisesMessage(
                ImproperlyConfigured,
                "RS256 alg requires OIDC_RP_IDP_SIGN_KEY or OIDC_OP_JWKS_ENDPOINT to be configured."):
            DjangocmsOIDCAuthenticationBackend()

    def test_verify_claims(self):
        retval = DjangocmsOIDCAuthenticationBackend().verify_claims({'email': 'email@example.com'})
        self.assertTrue(retval)

    def test_verify_claims_failed(self):
        retval = DjangocmsOIDCAuthenticationBackend().verify_claims({})
        self.assertFalse(retval)

    @override_settings(OIDC_RP_SCOPES="")
    def test_verify_claims_no_scopes(self):
        retval = DjangocmsOIDCAuthenticationBackend().verify_claims({})
        self.assertTrue(retval)


class TestDjangocmsOIDCAuthenticationBackend(TestCase):

    @classmethod
    def setUpTestData(cls):
        cls.provider = OIDCProvider.objects.create(
            name="Provider", slug="provider", client_id="example_id", client_secret="client_secret",
            token_endpoint="https://foo.foo/token", user_endpoint="https://foo.foo/user")
        cls.plugin = OIDCHandoverData.objects.create(provider=cls.provider, claims={})

    @override_settings(OIDC_OP_TOKEN_ENDPOINT='https://server.example.com/token')
    @override_settings(OIDC_OP_USER_ENDPOINT='https://server.example.com/user')
    @override_settings(OIDC_RP_CLIENT_ID='example_id')
    @override_settings(OIDC_RP_CLIENT_SECRET='client_secret')
    def setUp(self):
        request = RequestFactory().request()
        request.session = {}
        self.backend = DjangocmsOIDCAuthenticationBackend()
        self.backend.request = request

    @override_settings(OIDC_USE_NONCE=True)
    @patch('mozilla_django_oidc.auth.OIDCAuthenticationBackend._verify_jws')
    def test_verify_token_failed_nonce(self, jws_mock):
        self.backend.request.session[DJANGOCMS_PLUGIN_SESSION_KEY] = (self.plugin.consumer_type, self.plugin.pk)
        jws_mock.return_value = json.dumps({'nonce': 'foobar'}).encode('utf-8')
        with self.assertRaisesMessage(SuspiciousOperation, 'JWT Nonce verification failed.'):
            self.backend.verify_token('my_token', **{'nonce': 'foo'})
        jws_mock.assert_called_with(b'my_token', 'client_secret')

    @override_settings(OIDC_USE_NONCE=True)
    @patch('mozilla_django_oidc.auth.OIDCAuthenticationBackend._verify_jws')
    def test_verify_token(self, jws_mock):
        self.backend.request.session[DJANGOCMS_PLUGIN_SESSION_KEY] = (self.plugin.consumer_type, self.plugin.pk)
        jws_mock.return_value = json.dumps({'nonce': 'foobar'}).encode('utf-8')
        payload = self.backend.verify_token('my_token', **{'nonce': 'foobar'})
        jws_mock.assert_called_with(b'my_token', 'client_secret')
        self.assertEqual(payload, {'nonce': 'foobar'})

    def test_verify_token_no_consumer(self):
        payload = self.backend.verify_token('my_token', **{'nonce': 'foobar'})
        self.assertEqual(payload, {})

    def test_get_token_no_consumer(self):
        response = self.backend.get_token({})
        self.assertEqual(response, {})

    @requests_mock.Mocker()
    def test_get_token(self, mock_req):
        mock_req.post("https://foo.foo/token", json={'status': 'OK'})
        self.backend.request.session[DJANGOCMS_PLUGIN_SESSION_KEY] = (self.plugin.consumer_type, self.plugin.pk)
        response = self.backend.get_token({})
        self.assertEqual(response, {'status': 'OK'})

    @requests_mock.Mocker()
    def test_get_token_debug_response(self, mock_req):
        mock_req.post("https://foo.foo/token", json={'status': 'OK'})
        claims = {
            'userinfo': {'email': {'essential': True}},
            'debug_response': True,
        }
        plugin = OIDCHandoverData.objects.create(provider=self.provider, claims=claims)
        self.backend.request.session[DJANGOCMS_PLUGIN_SESSION_KEY] = (plugin.consumer_type, plugin.pk)
        self.backend.request._messages = FallbackStorage(self.backend.request)
        response = self.backend.get_token({})
        self.assertEqual(response, {'status': 'OK'})
        self.assertEqual(self._get_messages(self.backend.request), [
            (messages.WARNING,
                '<div>POST: https://foo.foo/token</div>'
                '<div>Payload: {}</div>'
                '<div>Auth: None</div>'
                '<div>Verify: True</div>'
                '<div>Timeout: None</div><div>Proxies: None</div>'),
            (messages.INFO, 'Status code: 200'),
            (messages.SUCCESS, 'Reason: None'),
            (messages.SUCCESS, 'b\'{"status": "OK"}\''),
        ])

    @override_settings(OIDC_TOKEN_USE_BASIC_AUTH=True)
    def test_get_token_use_basic_auth(self):
        response = self.backend.get_token({'client_id': '42', 'client_secret': 'secret'})
        self.assertEqual(response, {})

    def test_get_userinfo_no_consumer(self):
        response = self.backend.get_userinfo('access_token', 'id_token', {})
        self.assertEqual(response, {})

    @requests_mock.Mocker()
    def test_get_userinfo(self, mock_req):
        mock_req.get("https://foo.foo/user", json={'status': 'OK'})
        self.backend.request.session[DJANGOCMS_PLUGIN_SESSION_KEY] = (self.plugin.consumer_type, self.plugin.pk)
        response = self.backend.get_userinfo('access_token', 'id_token', {})
        self.assertEqual(response, {'status': 'OK'})

    def test_authenticate_no_request(self):
        self.assertIsNone(self.backend.authenticate(None))

    @requests_mock.Mocker()
    @patch('mozilla_django_oidc.auth.OIDCAuthenticationBackend._verify_jws')
    def test_authenticate(self, mock_req, jws_mock):
        jws_mock.return_value = json.dumps({'nonce': 'foobar'}).encode('utf-8')
        mock_req.post("https://foo.foo/token", json={'id_token': '42', 'access_token': 'ok'})
        mock_req.get("https://foo.foo/user", json={'email': 'user@foo.foo'})
        request = RequestFactory().get("/?state=ok&code=42")
        request.session = {
            DJANGOCMS_PLUGIN_SESSION_KEY: (self.plugin.consumer_type, self.plugin.pk)
        }
        self.backend.request = request
        self.assertIsNone(self.backend.authenticate(request, nonce='foobar'))
        self.assertEqual(request.session[DJANGOCMS_USER_SESSION_KEY]['email'], 'user@foo.foo')

    def test_authenticate_no_code(self):
        request = RequestFactory().get("/?state=ok")
        self.backend.request = request
        self.assertIsNone(self.backend.authenticate(request))

    @requests_mock.Mocker()
    @patch("djangocms_oidc.auth.DjangocmsOIDCAuthenticationBackend.verify_token")
    def test_authenticate_no_payload(self, mock_req, mock_verify_token):
        mock_req.post("https://foo.foo/token", json={'id_token': '42', 'access_token': 'ok'})
        mock_verify_token.return_value = None
        request = RequestFactory().get("/?state=ok&code=42")
        request.session = {
            DJANGOCMS_PLUGIN_SESSION_KEY: (self.plugin.consumer_type, self.plugin.pk)
        }
        self.backend.request = request
        self.assertIsNone(self.backend.authenticate(request, nonce='foobar'))
        mock_verify_token.assert_called_with('42', nonce='foobar')

    @requests_mock.Mocker()
    def test_authenticate_raise_for_status(self, mock_req):
        mock_req.post("https://foo.foo/token", exc=HTTPError("401 Client Error"))
        request = RequestFactory().get("/?state=ok&code=42")
        request.session = {
            DJANGOCMS_PLUGIN_SESSION_KEY: (self.plugin.consumer_type, self.plugin.pk)
        }
        self.backend.request = request
        self.backend.request._messages = FallbackStorage(self.backend.request)
        self.assertIsNone(self.backend.authenticate(request, nonce='foobar'))
        self.assertEqual(self._get_messages(self.backend.request), [
            (messages.ERROR, '401 Client Error')
        ])

    @requests_mock.Mocker()
    @patch("djangocms_oidc.auth.DjangocmsOIDCAuthenticationBackend.verify_token")
    def test_authenticate_no_plugin(self, mock_req, mock_verify_token):
        mock_req.post("https://foo.foo/token", json={'id_token': '42', 'access_token': 'ok'})
        mock_verify_token.return_value = None
        request = RequestFactory().get("/?state=ok&code=42")
        request.session = {}
        self.backend.request = request
        self.backend.request._messages = FallbackStorage(self.backend.request)
        self.assertIsNone(self.backend.authenticate(request, nonce='foobar'))
        mock_verify_token.assert_not_called()
        self.assertEqual(self._get_messages(self.backend.request), [
            (messages.ERROR, 'OIDC Consumer activation failed during authentication. Please try again.')
        ])

    @requests_mock.Mocker()
    @patch("djangocms_oidc.auth.DjangocmsOIDCAuthenticationBackend.verify_token")
    def test_authenticate_suspicious_operation(self, mock_req, mock_verify_token):
        mock_req.post("https://foo.foo/token", json={'id_token': '42', 'access_token': 'ok'})
        mock_req.get("https://foo.foo/user", exc=SuspiciousOperation)
        mock_verify_token.return_value = {'status': 'foo'}
        request = RequestFactory().get("/?state=ok&code=42")
        request.session = {
            DJANGOCMS_PLUGIN_SESSION_KEY: (self.plugin.consumer_type, self.plugin.pk)
        }
        self.backend.request = request
        self.assertIsNone(self.backend.authenticate(request, nonce='foobar'))
        mock_verify_token.assert_called_with('42', nonce='foobar')

    @requests_mock.Mocker()
    @patch("mozilla_django_oidc.auth.OIDCAuthenticationBackend.get_payload_data")
    def test_authenticate_suspicious_operation_verify_token(self, mock_req, mock_get_payload_data):
        mock_req.post("https://foo.foo/token", json={'id_token': '42', 'access_token': 'ok'})
        mock_get_payload_data.return_value = b'{"nonce": "unexpected_secret"}'
        request = RequestFactory().get("/?state=ok&code=42")
        request.session = {
            DJANGOCMS_PLUGIN_SESSION_KEY: (self.plugin.consumer_type, self.plugin.pk)
        }
        self.backend.request = request
        self.backend.request._messages = FallbackStorage(self.backend.request)
        self.assertIsNone(self.backend.authenticate(request, nonce='client_secret'))
        mock_get_payload_data.assert_called_with(b'42', 'client_secret')
        self.assertEqual(self._get_messages(self.backend.request), [
            (messages.ERROR, 'JWT Nonce verification failed.')
        ])

    @override_settings(OIDC_RP_SIGN_ALGO='RS256')
    @override_settings(OIDC_RP_IDP_SIGN_KEY='key')
    @patch("djangocms_oidc.auth.DjangocmsOIDCAuthenticationBackend.get_payload_data")
    def test_verify_token_algo_sign_key(self, mock_get_payload_data):
        mock_get_payload_data.return_value = json.dumps({'nonce': 'foobar'}).encode('utf-8')
        backend = DjangocmsOIDCAuthenticationBackend()
        backend.request = self.backend.request
        response = backend.verify_token('1234567980', nonce="foobar")
        self.assertEqual(response, {'nonce': 'foobar'})
        mock_get_payload_data.assert_called_with(b'1234567980', 'key')

    @override_settings(OIDC_RP_SIGN_ALGO='RS256')
    @override_settings(OIDC_RP_IDP_SIGN_KEY=None)
    @override_settings(OIDC_OP_JWKS_ENDPOINT="https://foo.foo/jwk-endpoint")
    @patch("djangocms_oidc.auth.DjangocmsOIDCAuthenticationBackend.retrieve_matching_jwk")
    @patch("djangocms_oidc.auth.DjangocmsOIDCAuthenticationBackend.get_payload_data")
    def test_verify_token_algo_sign_key_is_none(self, mock_get_payload_data, mock_retrieve_matching_jwk):
        mock_get_payload_data.return_value = json.dumps({'nonce': 'foobar'}).encode('utf-8')
        mock_retrieve_matching_jwk.return_value = 'key'
        backend = DjangocmsOIDCAuthenticationBackend()
        backend.request = self.backend.request
        response = backend.verify_token('1234567980', nonce="foobar")
        self.assertEqual(response, {'nonce': 'foobar'})
        mock_get_payload_data.assert_called_with(b'1234567980', 'key')
        mock_retrieve_matching_jwk.assert_called_with(b'1234567980')

    @requests_mock.Mocker()
    @patch("djangocms_oidc.auth.DjangocmsOIDCAuthenticationBackend.verify_claims")
    def test_get_or_create_user_no_claims_verified(self, mock_req, mock_verify_claims):
        mock_req.get("https://foo.foo/user", json={'status': 'OK'})
        mock_verify_claims.return_value = False
        self.backend.request.session[DJANGOCMS_PLUGIN_SESSION_KEY] = (self.plugin.consumer_type, self.plugin.pk)
        with self.assertRaisesMessage(SuspiciousOperation, "Claims verification failed"):
            self.backend.get_or_create_user(self.backend.request, 'access_token', 'id_token', {'status': 'foo'})
        mock_verify_claims.assert_called_with({'status': 'OK'})

    def _get_messages(self, request):
        return [(msg.level, str(msg.message)) for msg in get_messages(request)]

    @requests_mock.Mocker()
    @patch("djangocms_oidc.helpers.get_consumer")
    @patch("djangocms_oidc.auth.DjangocmsOIDCAuthenticationBackend.verify_claims")
    def test_get_or_create_user_consumer_is_none(self, mock_req, mock_verify_claims, mock_get_consumer):
        mock_req.get("https://foo.foo/user", json={'status': 'OK'})
        mock_verify_claims.return_value = True
        mock_get_consumer.side_effect = [(self.plugin.consumer_type, self.plugin.pk), None]
        self.backend.request._messages = FallbackStorage(self.backend.request)
        response = self.backend.get_or_create_user(self.backend.request, 'access_token', 'id_token', {'status': 'foo'})
        self.assertIsNone(response)
        mock_verify_claims.assert_called_with({})
        self.assertEqual(self._get_messages(self.backend.request), [
            (messages.ERROR, 'OIDC Consumer activation failed during authentication. Please try again.')
        ])

    @requests_mock.Mocker()
    @patch("djangocms_oidc.auth.DjangocmsOIDCAuthenticationBackend.verify_claims")
    def test_get_or_create_user_consumer_insist(self, mock_req, mock_verify_claims):
        mock_req.get("https://foo.foo/user", json={'status': 'OK'})
        mock_verify_claims.return_value = True
        claims = {
            'userinfo': {'email': {'essential': True}},
        }
        plugin = OIDCHandoverData.objects.create(provider=self.provider, insist_on_required_claims=True, claims=claims)
        self.backend.request.session[DJANGOCMS_PLUGIN_SESSION_KEY] = (plugin.consumer_type, plugin.pk)
        self.backend.request._messages = FallbackStorage(self.backend.request)
        response = self.backend.get_or_create_user(self.backend.request, 'access_token', 'id_token', {'status': 'foo'})
        self.assertIsNone(response)
        mock_verify_claims.assert_called_with({'status': 'OK'})
        self.assertEqual(self._get_messages(self.backend.request), [
            (messages.ERROR, 'Not all required data has been handovered. Please make the transfer again and select all'
                             ' the required values.')
        ])

    @freeze_time('2020-10-12 18:55:42')
    @requests_mock.Mocker()
    @patch("djangocms_oidc.auth.DjangocmsOIDCAuthenticationBackend.verify_claims")
    def test_get_or_create_user_consumer_can_login(self, mock_req, mock_verify_claims):
        mock_req.get("https://foo.foo/user", json={'email': 'foo@foo.foo'})
        mock_verify_claims.return_value = True
        plugin = OIDCLogin.objects.create(provider=self.provider, insist_on_required_claims=True, claims={})
        self.backend.request.session[DJANGOCMS_PLUGIN_SESSION_KEY] = (plugin.consumer_type, plugin.pk)
        self.backend.request._messages = FallbackStorage(self.backend.request)
        self.backend.request.user = AnonymousUser()
        user = get_user_model().objects.create(username="user", email='foo@foo.foo')
        response = self.backend.get_or_create_user(self.backend.request, 'access_token', 'id_token', {'status': 'foo'})
        self.assertEqual(response, user)
        mock_verify_claims.assert_called_with({'email': 'foo@foo.foo', 'user_info_created_at': 1602528942.0})
        self.assertFalse(OIDCIdentifier.objects.filter(user=user).exists())
        self.assertEqual(self._get_messages(self.backend.request), [])

    @freeze_time('2020-10-12 18:55:42')
    @requests_mock.Mocker()
    @patch("djangocms_oidc.auth.DjangocmsOIDCAuthenticationBackend.verify_claims")
    def test_get_or_create_user_is_authenticated(self, mock_req, mock_verify_claims):
        mock_req.get("https://foo.foo/user", json={'email': 'foo@foo.foo'})
        mock_verify_claims.return_value = True
        plugin = OIDCLogin.objects.create(provider=self.provider, insist_on_required_claims=True, claims={})
        self.backend.request.session[DJANGOCMS_PLUGIN_SESSION_KEY] = (plugin.consumer_type, plugin.pk)
        self.backend.request._messages = FallbackStorage(self.backend.request)
        user = get_user_model().objects.create(username="user", email='foo@foo.foo')
        self.backend.request.user = user
        response = self.backend.get_or_create_user(self.backend.request, 'access_token', 'id_token', {'status': 'foo'})
        self.assertEqual(response, user)
        mock_verify_claims.assert_called_with({'email': 'foo@foo.foo', 'user_info_created_at': 1602528942.0})
        self.assertFalse(OIDCIdentifier.objects.filter(user=user).exists())
        self.assertEqual(self._get_messages(self.backend.request), [])

    @freeze_time('2020-10-12 18:55:42')
    @requests_mock.Mocker()
    @patch("djangocms_oidc.auth.DjangocmsOIDCAuthenticationBackend.verify_claims")
    def test_get_or_create_user_is_authenticated_with_openid2_id(self, mock_req, mock_verify_claims):
        mock_req.get("https://foo.foo/user", json={'email': 'foo@foo.foo', 'openid2_id': 'oid'})
        mock_verify_claims.return_value = True
        plugin = OIDCLogin.objects.create(provider=self.provider, insist_on_required_claims=True, claims={})
        self.backend.request.session[DJANGOCMS_PLUGIN_SESSION_KEY] = (plugin.consumer_type, plugin.pk)
        self.backend.request._messages = FallbackStorage(self.backend.request)
        user = get_user_model().objects.create(username="user", email='foo@foo.foo')
        self.backend.request.user = user
        response = self.backend.get_or_create_user(self.backend.request, 'access_token', 'id_token', {'status': 'foo'})
        self.assertEqual(response, user)
        mock_verify_claims.assert_called_with({
            'email': 'foo@foo.foo', 'openid2_id': 'oid', 'user_info_created_at': 1602528942.0})
        self.assertQuerysetEqual(OIDCIdentifier.objects.all().values_list('uident', flat=True), ['oid'], transform=str)
        self.assertEqual(self._get_messages(self.backend.request), [
            (messages.SUCCESS, 'The account has been successfully paired with the provider.')
        ])

    @freeze_time('2020-10-12 18:55:42')
    @requests_mock.Mocker()
    @patch("djangocms_oidc.auth.DjangocmsOIDCAuthenticationBackend.verify_claims")
    def test_get_or_create_user_is_authenticated_openid2_id_already_paired(self, mock_req, mock_verify_claims):
        mock_req.get("https://foo.foo/user", json={'email': 'foo@foo.foo', 'openid2_id': 'oid'})
        mock_verify_claims.return_value = True
        plugin = OIDCLogin.objects.create(provider=self.provider, insist_on_required_claims=True, claims={})
        self.backend.request.session[DJANGOCMS_PLUGIN_SESSION_KEY] = (plugin.consumer_type, plugin.pk)
        self.backend.request._messages = FallbackStorage(self.backend.request)
        user = get_user_model().objects.create(username="user", email='foo@foo.foo')
        self.backend.request.user = user
        other_user = get_user_model().objects.create(username="other")
        OIDCIdentifier.objects.create(user=other_user, provider=self.provider, uident="oid")
        response = self.backend.get_or_create_user(self.backend.request, 'access_token', 'id_token', {'status': 'foo'})
        self.assertEqual(response, user)
        mock_verify_claims.assert_called_with({
            'email': 'foo@foo.foo', 'openid2_id': 'oid', 'user_info_created_at': 1602528942.0})
        self.assertFalse(OIDCIdentifier.objects.filter(user=user).exists())
        self.assertEqual(self._get_messages(self.backend.request), [
            (messages.ERROR, 'Pairing cannot be performed. This identifier is already paired with another account.')
        ])

    def test_create_identifier_if_missing(self):
        openid2_id = "openid2_id"
        user = get_user_model().objects.create(username="user")
        OIDCIdentifier.objects.create(user=user, provider=self.provider, uident=openid2_id)
        self.backend.create_identifier_if_missing(self.backend.request, user, self.provider, openid2_id)
        self.assertQuerysetEqual(OIDCIdentifier.objects.all().values_list('uident', flat=True), [openid2_id],
                                 transform=str)

    def test_not_authenticated_user_zero(self):
        user = get_user_model().objects.create(username="user", email='foo@foo.foo')
        user_info = {'email': 'unknown@foo.foo'}
        self.backend.request._messages = FallbackStorage(self.backend.request)
        response = self.backend.not_authenticated_user(self.backend.request, user_info, self.plugin)
        self.assertIsNone(response)
        self.assertFalse(OIDCIdentifier.objects.filter(user=user).exists())
        self.assertEqual(self._get_messages(self.backend.request), [])

    def test_not_authenticated_user_zero_with_openid2_id(self):
        user = get_user_model().objects.create(username="user", email='foo@foo.foo')
        user_info = {'email': 'unknown@foo.foo', 'openid2_id': '4242'}
        self.backend.request._messages = FallbackStorage(self.backend.request)
        response = self.backend.not_authenticated_user(self.backend.request, user_info, self.plugin)
        self.assertIsNone(response)
        self.assertFalse(OIDCIdentifier.objects.filter(user=user).exists())
        self.assertEqual(self._get_messages(self.backend.request), [])

    def test_not_authenticated_user_one(self):
        user = get_user_model().objects.create(username="user", email='foo@foo.foo')
        user_info = {'email': user.email}
        self.backend.request._messages = FallbackStorage(self.backend.request)
        response = self.backend.not_authenticated_user(self.backend.request, user_info, self.plugin)
        self.assertEqual(response, user)
        self.assertFalse(OIDCIdentifier.objects.filter(user=user).exists())
        self.assertEqual(self._get_messages(self.backend.request), [])

    def test_not_authenticated_user_one_with_openid2_id(self):
        user = get_user_model().objects.create(username="user", email='foo@foo.foo')
        user_info = {'email': user.email, 'openid2_id': '4242'}
        self.backend.request._messages = FallbackStorage(self.backend.request)
        response = self.backend.not_authenticated_user(self.backend.request, user_info, self.plugin)
        self.assertEqual(response, user)
        self.assertQuerysetEqual(OIDCIdentifier.objects.filter(user=user).values_list('uident', flat=True), ['4242'],
                                 transform=str)
        self.assertEqual(self._get_messages(self.backend.request), [
            (messages.SUCCESS, 'The account has been successfully paired with the provider.')
        ])

    def test_not_authenticated_user_one_user_is_not_active(self):
        user = get_user_model().objects.create(username="user", email='foo@foo.foo', is_active=False)
        user_info = {'email': user.email, 'openid2_id': '4242'}
        self.backend.request._messages = FallbackStorage(self.backend.request)
        response = self.backend.not_authenticated_user(self.backend.request, user_info, self.plugin)
        self.assertIsNone(response)
        self.assertQuerysetEqual(OIDCIdentifier.objects.filter(user=user).values_list('uident', flat=True), ['4242'],
                                 transform=str)
        self.assertEqual(self._get_messages(self.backend.request), [
            (messages.SUCCESS, 'The account has been successfully paired with the provider.'),
            (messages.ERROR, 'Your account is deactivated. Please contact our support.'),
        ])

    def test_not_authenticated_user_more_than_one(self):
        User = get_user_model()
        User.objects.create(username="user1", email='foo@foo.foo')
        User.objects.create(username="user2", email='foo@foo.foo')
        user_info = {'email': 'foo@foo.foo'}
        self.backend.request._messages = FallbackStorage(self.backend.request)
        with self.assertRaisesMessage(SuspiciousOperation, 'Multiple users returned'):
            self.backend.not_authenticated_user(self.backend.request, user_info, self.plugin)

    def test_not_authenticated_user_can_login(self):
        plugin = OIDCLogin.objects.create(provider=self.provider, claims={})
        user_info = {'email': 'foo@foo.foo'}
        self.backend.request._messages = FallbackStorage(self.backend.request)
        response = self.backend.not_authenticated_user(self.backend.request, user_info, plugin)
        user = get_user_model().objects.get(email='foo@foo.foo')
        self.assertEqual(response, user)
        self.assertQuerysetEqual(OIDCIdentifier.objects.all(), [])
        self.assertEqual(self._get_messages(self.backend.request), [
            (messages.SUCCESS, 'A new account has been created with the username '
                               'aLOwvSRJuFVqFX111xmv5vYGuXk and email foo@foo.foo.'),
        ])

    def test_not_authenticated_user_can_login_with_openid2_id(self):
        plugin = OIDCLogin.objects.create(provider=self.provider, claims={})
        user_info = {'email': 'foo@foo.foo', 'openid2_id': '4242'}
        self.backend.request._messages = FallbackStorage(self.backend.request)
        response = self.backend.not_authenticated_user(self.backend.request, user_info, plugin)
        user = get_user_model().objects.get(email='foo@foo.foo')
        self.assertEqual(response, user)
        self.assertQuerysetEqual(OIDCIdentifier.objects.filter(user=user).values_list('uident', flat=True), ['4242'],
                                 transform=str)
        self.assertEqual(self._get_messages(self.backend.request), [
            (messages.SUCCESS, 'A new account has been created with the username '
                               'aLOwvSRJuFVqFX111xmv5vYGuXk and email foo@foo.foo.'),
        ])

    def test_not_authenticated_user_can_login_no_new_user(self):
        plugin = OIDCLogin.objects.create(provider=self.provider, no_new_user=True, claims={})
        user_info = {'email': 'foo@foo.foo'}
        self.backend.request._messages = FallbackStorage(self.backend.request)
        response = self.backend.not_authenticated_user(self.backend.request, user_info, plugin)
        self.assertIsNone(response)
        self.assertEqual(self._get_messages(self.backend.request), [
            (messages.INFO, 'To pair with your identity provider, log in first.'),
        ])



from django import forms
from django.utils.translation import gettext_lazy as _


class OIDCProviderForm(forms.ModelForm):

    def clean(self):
        cleaned_data = super().clean()
        if cleaned_data.get('register_consumer') is None and (
                cleaned_data.get('client_id') is None or cleaned_data.get('client_secret') is None):
            self.add_error(
                'client_id', forms.ValidationError(_('The value is required if "Register consumer" is not set.')))
            self.add_error(
                'client_secret', forms.ValidationError(_('The value is required if "Register consumer" is not set.')))



# Generated by Django 2.2.13 on 2020-09-04 14:59

import cms.models.fields
from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion
import djangocms_oidc.models


class Migration(migrations.Migration):

    initial = True

    dependencies = []

    operations = [
        migrations.CreateModel(
            name='OIDCDisplayDedicatedContent',
            fields=[
                ('cmsplugin_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, related_name='djangocms_oidc_oidcdisplaydedicatedcontent', serialize=False, to='cms.CMSPlugin')),
                ('conditions', models.CharField(blank=True, choices=[('email_verified', 'Email verified')], help_text='Show content only if the given conditions are met.', max_length=255, null=True, verbose_name='Conditions')),
            ],
            options={
                'abstract': False,
            },
            bases=('cms.cmsplugin',),
        ),
        migrations.CreateModel(
            name='OIDCRegisterConsumer',
            fields=[
                ('cmsplugin_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, related_name='djangocms_oidc_oidcregisterconsumer', serialize=False, to='cms.CMSPlugin')),
                ('name', models.CharField(default='Default', help_text='Registration name', max_length=255, unique=True, verbose_name='Name')),
                ('register_url', models.URLField(help_text='Register consumer automaticaly, when it is not. (Client ID and secret missing).', verbose_name='Register consumer URL')),
            ],
            options={
                'abstract': False,
            },
            bases=('cms.cmsplugin',),
        ),
        migrations.CreateModel(
            name='OIDCShowAttribute',
            fields=[
                ('cmsplugin_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, related_name='djangocms_oidc_oidcshowattribute', serialize=False, to='cms.CMSPlugin')),
                ('verified_by', models.CharField(default='name', help_text='Verified by names (separated by space). Eg.: name given_name+family_name preferred_username nickname email openid2_id', max_length=255, verbose_name='Verified by names')),
                ('default_value', models.CharField(blank=True, help_text='Show this text when attribute has not been handovered.', max_length=255, null=True, verbose_name='Default value')),
                ('css_names', models.CharField(blank=True, help_text='Extra class names (separated by space).', max_length=255, null=True, verbose_name='CSS names')),
            ],
            options={
                'abstract': False,
            },
            bases=('cms.cmsplugin',),
        ),
        migrations.CreateModel(
            name='OIDCProvider',
            fields=[
                ('cmsplugin_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, related_name='djangocms_oidc_oidcprovider', serialize=False, to='cms.CMSPlugin')),
                ('name', models.CharField(help_text='Provider name', max_length=255, unique=True, verbose_name='Name')),
                ('slug', models.SlugField(help_text='Used, for example, as a css class for a button or link.', verbose_name='Code')),
                ('client_id', models.CharField(blank=True, help_text='The value is required if "Register consumer" is not set.', max_length=255, null=True, verbose_name='Client ID')),
                ('client_secret', models.CharField(blank=True, help_text='The value is required if "Register consumer" is not set.', max_length=255, null=True, verbose_name='Client secret')),
                ('authorization_endpoint', models.URLField(verbose_name='Authorization endpoint')),
                ('token_endpoint', models.URLField(verbose_name='Token endpoint')),
                ('user_endpoint', models.URLField(verbose_name='User endpoint')),
                ('account_url', models.URLField(blank=True, null=True, verbose_name='Account URL')),
                ('logout_url', models.URLField(blank=True, null=True, verbose_name='Logout URL')),
                ('register_consumer', models.ForeignKey(blank=True, help_text='Register consumer on the provider, when Client ID and secret are not set.', null=True, on_delete=django.db.models.deletion.CASCADE, to='djangocms_oidc.OIDCRegisterConsumer', verbose_name='Register consumer')),
            ],
            options={
                'abstract': False,
            },
            bases=('cms.cmsplugin',),
        ),
        migrations.CreateModel(
            name='OIDCLogin',
            fields=[
                ('cmsplugin_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, related_name='djangocms_oidc_oidclogin', serialize=False, to='cms.CMSPlugin')),
                ('claims', models.JSONField(default=dict, help_text='Claims attributes for data handover.', validators=[djangocms_oidc.models.validate_claims], verbose_name='Claims')),
                ('insist_on_required_claims', models.BooleanField(default=False, help_text='Consider the data invalid if not all the required data has been handovered.', verbose_name='Insist on required claims')),
                ('verified_by', models.CharField(blank=True, help_text='Verified by names (separated by space).', max_length=255, null=True, verbose_name='Verified by names')),
                ('authorization_prompt', models.CharField(blank=True, choices=[('none', 'No interaction'), ('login', 'Force login'), ('consent', 'Force consent with handovered data')], help_text='Prompt for user at authorization. "No interaction" cannot be combined with others.', max_length=255, null=True, verbose_name='Prompt')),
                ('no_new_user', models.BooleanField(default=False, help_text='Do not create a new user account if the user is not recognized.', verbose_name='No new user')),
                ('provider', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='djangocms_oidc.OIDCProvider')),
                ('redirect_page', cms.models.fields.PageField(blank=True, help_text='Redirect to the page after login or hangover data.', null=True, on_delete=django.db.models.deletion.SET_NULL, to='cms.Page', verbose_name='CMS Page')),
            ],
            options={
                'abstract': False,
            },
            bases=('cms.cmsplugin',),
        ),
        migrations.CreateModel(
            name='OIDCIdentifier',
            fields=[
                ('cmsplugin_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, related_name='djangocms_oidc_oidcidentifier', serialize=False, to='cms.CMSPlugin')),
                ('uident', models.CharField(help_text='Provider unique identifier.', max_length=255, unique=True, verbose_name='Identifier')),
                ('provider', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='djangocms_oidc.OIDCProvider')),
                ('user', models.OneToOneField(on_delete=django.db.models.deletion.CASCADE, to=settings.AUTH_USER_MODEL)),
            ],
            options={
                'abstract': False,
            },
            bases=('cms.cmsplugin',),
        ),
        migrations.CreateModel(
            name='OIDCHandoverData',
            fields=[
                ('cmsplugin_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, related_name='djangocms_oidc_oidchandoverdata', serialize=False, to='cms.CMSPlugin')),
                ('claims', models.JSONField(default=dict, help_text='Claims attributes for data handover.', validators=[djangocms_oidc.models.validate_claims], verbose_name='Claims')),
                ('insist_on_required_claims', models.BooleanField(default=False, help_text='Consider the data invalid if not all the required data has been handovered.', verbose_name='Insist on required claims')),
                ('verified_by', models.CharField(blank=True, help_text='Verified by names (separated by space).', max_length=255, null=True, verbose_name='Verified by names')),
                ('authorization_prompt', models.CharField(blank=True, choices=[('none', 'No interaction'), ('login', 'Force login'), ('consent', 'Force consent with handovered data')], help_text='Prompt for user at authorization. "No interaction" cannot be combined with others.', max_length=255, null=True, verbose_name='Prompt')),
                ('provider', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='djangocms_oidc.OIDCProvider')),
                ('redirect_page', cms.models.fields.PageField(blank=True, help_text='Redirect to the page after login or hangover data.', null=True, on_delete=django.db.models.deletion.SET_NULL, to='cms.Page', verbose_name='CMS Page')),
            ],
            options={
                'abstract': False,
            },
            bases=('cms.cmsplugin',),
        ),
    ]



from django import forms
from django.core.exceptions import ValidationError
from django.utils.translation import gettext_lazy as _

from .widgets import JsonDataTextarea


class OIDCDataForm(forms.ModelForm):

    class Meta:
        widgets = {
            "claims": JsonDataTextarea,
        }

    def clean_authorization_prompt(self):
        prompt = self.cleaned_data.get('authorization_prompt')
        if "none" in prompt and len(prompt) > 1:
            raise ValidationError(_('Item "No interaction" cannot be combined with others.'))
        return prompt


class OIDCHandoverDataForm(forms.ModelForm):

    class Meta:
        widgets = {
            "claims": JsonDataTextarea,
        }



from cms.views import details
from django.urls import path, re_path
from mozilla_django_oidc.urls import urlpatterns as mozilla_urlpatterns

from djangocms_oidc.tests.views import TestFakePage, TestHomePage, TestLoginPage
from djangocms_oidc.urls import urlpatterns as djangocms_oidc_urlpatterns

urlpatterns = [
    path('', TestHomePage.as_view(), name='test_home_page'),
    path('login/', TestLoginPage.as_view(), name='login'),
    path('mdo_fake_view/', TestFakePage.as_view(), name='mdo_fake_view'),
]
urlpatterns.extend(mozilla_urlpatterns)
urlpatterns.extend(djangocms_oidc_urlpatterns)
urlpatterns.append(re_path(r'^(?P<slug>[0-9A-Za-z-_.//]+)/$', details, name='pages-details-by-slug'))



from django.forms import modelform_factory
from django.test import SimpleTestCase

from djangocms_oidc.forms import OIDCDataForm
from djangocms_oidc.models import OIDCHandoverData


class TestOIDCDataForm(SimpleTestCase):

    def setUp(self):
        self.Form = modelform_factory(OIDCHandoverData, form=OIDCDataForm, fields=('authorization_prompt', ))

    def test_none_login(self):
        form = self.Form({'authorization_prompt': ['none', 'login']})
        self.assertEqual(form.errors, {
            'authorization_prompt': ['Item "No interaction" cannot be combined with others.']
        })

    def test_none(self):
        form = self.Form({'authorization_prompt': ['none']})
        self.assertTrue(form.is_valid())

    def test_consent_login(self):
        form = self.Form({'authorization_prompt': ['consent', 'login']})
        self.assertTrue(form.is_valid())



SECRET_KEY = "secret"

INSTALLED_APPS = [
    # Django
    "django.contrib.admin",
    "django.contrib.auth",
    "django.contrib.contenttypes",
    "django.contrib.messages",
    "django.contrib.sessions",
    "django.contrib.staticfiles",
    "django.contrib.sites",
    # Django CMS
    'cms',
    'menus',
    'treebeard',
    'sekizai',
    # dependencies
    'mozilla_django_oidc',
    'django_countries',
    # the project
    'djangocms_oidc',
]

AUTHENTICATION_BACKENDS = [
    'django.contrib.auth.backends.ModelBackend',
    'djangocms_oidc.auth.DjangocmsOIDCAuthenticationBackend',
]

MIDDLEWARE = [
    "django.middleware.security.SecurityMiddleware",
    "django.contrib.sessions.middleware.SessionMiddleware",
    "django.middleware.common.CommonMiddleware",
    "django.middleware.csrf.CsrfViewMiddleware",
    "django.contrib.auth.middleware.AuthenticationMiddleware",
    "django.contrib.messages.middleware.MessageMiddleware",
    "django.middleware.clickjacking.XFrameOptionsMiddleware",
    'cms.middleware.user.CurrentUserMiddleware',
    'cms.middleware.page.CurrentPageMiddleware',
    'cms.middleware.toolbar.ToolbarMiddleware',
    'cms.middleware.language.LanguageCookieMiddleware',
    'djangocms_oidc.middleware.OIDCSessionRefresh',
]

TEMPLATES = [
    {
        "BACKEND": "django.template.backends.django.DjangoTemplates",
        "DIRS": ['djangocms_oidc/tests'],
        "APP_DIRS": True,
        "OPTIONS": {
            "context_processors": [
                "django.template.context_processors.debug",
                "django.template.context_processors.request",
                "django.contrib.auth.context_processors.auth",
                "django.contrib.messages.context_processors.messages",
                'sekizai.context_processors.sekizai',
            ],
        },
    },
]

CMS_TEMPLATES = (
    ('test_content_plugin.html', 'Test content'),
)

DATABASES = {
    "default": {
        "ENGINE": "django.db.backends.sqlite3",
        "NAME": "tests",
    }
}

TIME_ZONE = "UTC"
USE_TZ = True

LANGUAGE_CODE = "en"
LANGUAGES = [
    ('en', 'English'),
    ('cs', 'Czech'),
]

SITE_ID = 1
ROOT_URLCONF = 'djangocms_oidc.tests.urls'

OIDC_AUTHENTICATE_CLASS = "djangocms_oidc.views.DjangocmsOIDCAuthenticationRequestView"
OIDC_CALLBACK_CLASS = "djangocms_oidc.views.DjangocmsOIDCAuthenticationCallbackView"

OIDC_OP_AUTHORIZATION_ENDPOINT = "https://foo.foo/authorization-endpoint"
OIDC_RP_CLIENT_ID = "cid"



from cms.utils.conf import get_cms_setting
from django.utils.encoding import force_str
from django.utils.translation import gettext_lazy as _
from mozilla_django_oidc.utils import import_from_settings


def get_settings(attr, *args):
    return import_from_settings(attr, *args)


def get_cache_key(key):
    return "{}{}".format(get_cms_setting('CACHE_PREFIX'), key)


def only_user_is_staff(context, instance, placeholder, user_info):
    request = context['request']
    return hasattr(request, "user") and request.user.is_staff


def only_authenticated_user(context, instance, placeholder, user_info):
    request = context['request']
    return hasattr(request, "user") and request.user.is_authenticated


def email_verified(context, instance, placeholder, user_info):
    if user_info is None:
        return False
    messages = []
    if not user_info.get('email'):
        messages.append(_("Email missing."))
    if not user_info.get('email_verified'):
        messages.append(_("Email is not verified."))
    if messages:
        context['dedicated_content'] = "<ul class='messagelist'><li class='error'>{}</li></ul>".format(
            " ".join([force_str(msg) for msg in messages]))
    return len(messages) == 0


DEFAULT_DISPLAY_CONTENT = [
    # (Choice code, choice label, function)
    ('only_authenticated_user', _('Only authenticated user'), only_authenticated_user),
    ('only_user_is_staff', _('Only user is staff'), only_user_is_staff),
    ('email_verified', _('Email verified'), email_verified),
]



from django.contrib import admin

from ..models import OIDCIdentifier, OIDCProvider, OIDCRegisterConsumer
from .options import OIDCIdentifierAdmin, OIDCProviderAdmin, OIDCRegisterConsumerAdmin

admin.site.register(OIDCRegisterConsumer, OIDCRegisterConsumerAdmin)
admin.site.register(OIDCProvider, OIDCProviderAdmin)
admin.site.register(OIDCIdentifier, OIDCIdentifierAdmin)



import json

import requests
from django.contrib import messages
from django.contrib.auth import get_user_model
from django.core.exceptions import ImproperlyConfigured
from django.urls import reverse
from django.utils import timezone
from django.utils.encoding import force_bytes
from django.utils.safestring import mark_safe
from django.utils.translation import gettext_lazy as _
from mozilla_django_oidc.auth import LOGGER, OIDCAuthenticationBackend, SuspiciousOperation
from mozilla_django_oidc.utils import absolutify
from requests.auth import HTTPBasicAuth

from .constants import DJANGOCMS_USER_SESSION_KEY
from .helpers import check_required_handovered, get_consumer
from .models import OIDCIdentifier


class DjangocmsOIDCAuthenticationBackend(OIDCAuthenticationBackend):
    """Provide OIDC backend with values taken from consumer plugin."""

    def __init__(self, *args, **kwargs):
        """Initialize settings."""
        self.OIDC_OP_TOKEN_ENDPOINT = self.get_settings('OIDC_OP_TOKEN_ENDPOINT', None)  # taken from plugin
        self.OIDC_OP_USER_ENDPOINT = self.get_settings('OIDC_OP_USER_ENDPOINT', None)  # taken from plugin
        self.OIDC_OP_JWKS_ENDPOINT = self.get_settings('OIDC_OP_JWKS_ENDPOINT', None)
        self.OIDC_RP_CLIENT_ID = self.get_settings('OIDC_RP_CLIENT_ID', None)  # taken from plugin
        self.OIDC_RP_CLIENT_SECRET = self.get_settings('OIDC_RP_CLIENT_SECRET', None)  # taken from plugin
        self.OIDC_RP_SIGN_ALGO = self.get_settings('OIDC_RP_SIGN_ALGO', 'HS256')
        self.OIDC_RP_IDP_SIGN_KEY = self.get_settings('OIDC_RP_IDP_SIGN_KEY', None)

        if (self.OIDC_RP_SIGN_ALGO.startswith('RS') and
                (self.OIDC_RP_IDP_SIGN_KEY is None and self.OIDC_OP_JWKS_ENDPOINT is None)):
            msg = '{} alg requires OIDC_RP_IDP_SIGN_KEY or OIDC_OP_JWKS_ENDPOINT to be configured.'
            raise ImproperlyConfigured(msg.format(self.OIDC_RP_SIGN_ALGO))

        self.UserModel = get_user_model()

    def verify_claims(self, claims):
        """Verify the provided claims to decide if authentication should be allowed."""
        scopes = self.get_settings('OIDC_RP_SCOPES', 'openid2_id openid email').split()
        if scopes:
            # At least one of the `scopes` set is present in the `claims`.
            return bool(set(scopes) & set(claims.keys()))
        return True

    def verify_token(self, token, **kwargs):
        """Validate the token signature."""
        nonce = kwargs.get('nonce')

        token = force_bytes(token)
        if self.OIDC_RP_SIGN_ALGO.startswith('RS'):
            if self.OIDC_RP_IDP_SIGN_KEY is not None:
                key = self.OIDC_RP_IDP_SIGN_KEY
            else:
                key = self.retrieve_matching_jwk(token)
        else:
            consumer = get_consumer(self.request)
            if consumer is None:
                return {}
            else:
                key = consumer.provider.get_client_secret()  # OIDC_RP_CLIENT_SECRET

        payload_data = self.get_payload_data(token, key)

        # The 'token' will always be a byte string since it's
        # the result of base64.urlsafe_b64decode().
        # The payload is always the result of base64.urlsafe_b64decode().
        # In Python 3 and 2, that's always a byte string.
        # In Python3.6, the json.loads() function can accept a byte string
        # as it will automagically decode it to a unicode string before
        # deserializing https://bugs.python.org/issue17909
        payload = json.loads(payload_data.decode('utf-8'))
        token_nonce = payload.get('nonce')

        if self.get_settings('OIDC_USE_NONCE', True) and nonce != token_nonce:
            msg = 'JWT Nonce verification failed.'
            raise SuspiciousOperation(msg)
        return payload

    def get_token(self, payload):
        """Return token object as a dictionary."""

        auth = None
        if self.get_settings('OIDC_TOKEN_USE_BASIC_AUTH', False):
            # When Basic auth is defined, create the Auth Header and remove secret from payload.
            user = payload.get('client_id')
            pw = payload.get('client_secret')

            auth = HTTPBasicAuth(user, pw)
            del payload['client_secret']

        consumer = get_consumer(self.request)
        if consumer is None:
            return {}
        response = requests.post(
            consumer.provider.token_endpoint,  # self.OIDC_OP_TOKEN_ENDPOINT,
            data=payload,
            auth=auth,
            verify=self.get_settings('OIDC_VERIFY_SSL', True),
            timeout=self.get_settings('OIDC_TIMEOUT', None),
            proxies=self.get_settings('OIDC_PROXY', None))
        # Debug request / response.
        if consumer.claims.get('debug_response', False):
            messages.warning(self.request, mark_safe(
                f"<div>POST: {consumer.provider.token_endpoint}</div>"
                f"<div>Payload: {payload}</div>"
                f"<div>Auth: {auth}</div>"
                f"<div>Verify: {self.get_settings('OIDC_VERIFY_SSL', True)}</div>"
                f"<div>Timeout: {self.get_settings('OIDC_TIMEOUT', None)}</div>"
                f"<div>Proxies: {self.get_settings('OIDC_PROXY', None)}</div>"
            ))
            messages.info(self.request, f"Status code: {response.status_code}")
            message_fnc = messages.success if response.status_code == 200 else messages.error
            message_fnc(self.request, f"Reason: {response.reason}")
            messages.info(self.request, mark_safe("<br/>\n".join(
                [f"{key}: {value}" for key, value in response.headers.items()])))
            message_fnc(self.request, response.content)
        response.raise_for_status()
        return response.json()

    def get_userinfo(self, access_token, id_token, payload):
        """Return user details dictionary. The id_token and payload are not used in
        the default implementation, but may be used when overriding this method"""

        consumer = get_consumer(self.request)
        if consumer is None:
            return {}
        user_response = requests.get(
            consumer.provider.user_endpoint,  # self.OIDC_OP_USER_ENDPOINT,
            headers={
                'Authorization': f'Bearer {access_token}'
            },
            verify=self.get_settings('OIDC_VERIFY_SSL', True),
            timeout=self.get_settings('OIDC_TIMEOUT', None),
            proxies=self.get_settings('OIDC_PROXY', None))
        user_response.raise_for_status()
        return user_response.json()

    # This is a copy of the class OIDCAuthenticationBackend with only the change in `get_or_create_user(request, ...)`.
    def authenticate(self, request, **kwargs):
        """Authenticates a user based on the OIDC code flow."""

        self.request = request
        if not self.request:
            return None

        state = self.request.GET.get('state')
        code = self.request.GET.get('code')
        nonce = kwargs.pop('nonce', None)

        if not code or not state:
            return None

        reverse_url = self.get_settings('OIDC_AUTHENTICATION_CALLBACK_URL',
                                        'oidc_authentication_callback')

        consumer = get_consumer(request)
        if consumer is None:
            messages.error(request, _('OIDC Consumer activation failed during authentication. Please try again.'))
            return None

        token_payload = {
            'client_id': consumer.provider.get_client_id(),  # replace OIDC_RP_CLIENT_ID
            'client_secret': consumer.provider.get_client_secret(),  # replace OIDC_RP_CLIENT_SECRET
            'grant_type': 'authorization_code',
            'code': code,
            'redirect_uri': absolutify(
                self.request,
                reverse(reverse_url)
            ),
        }

        # Get the token
        try:
            token_info = self.get_token(token_payload)
        except requests.exceptions.RequestException as msg:
            messages.error(request, msg)
            return None

        id_token = token_info.get('id_token')
        access_token = token_info.get('access_token')

        # Validate the token
        try:
            payload = self.verify_token(id_token, nonce=nonce)
        except SuspiciousOperation as msg:
            messages.error(request, msg)
            return None

        if payload:
            self.store_tokens(access_token, id_token)
            try:
                return self.get_or_create_user(request, access_token, id_token, payload)
            except SuspiciousOperation as exc:
                LOGGER.warning('failed to get or create user: %s', exc)
                return None

        return None

    def set_user_info_into_session(self, request, user_info):
        """Set user_info into session."""
        user_info['user_info_created_at'] = timezone.now().timestamp()
        self.request.session[DJANGOCMS_USER_SESSION_KEY] = user_info

    def get_or_create_user(self, request, access_token, id_token, payload):
        """Returns a User instance if 1 user is found. Creates a user if not found
        and configured to do so. Returns nothing if multiple users are matched."""

        user_info = self.get_userinfo(access_token, id_token, payload)

        claims_verified = self.verify_claims(user_info)
        if not claims_verified:
            msg = 'Claims verification failed'
            raise SuspiciousOperation(msg)

        consumer = get_consumer(request)
        if consumer is None:
            messages.error(request, _('OIDC Consumer activation failed during authentication. Please try again.'))
            return None

        if consumer.insist_on_required_claims and not check_required_handovered(consumer, user_info):
            messages.error(request, _('Not all required data has been handovered. '
                                      'Please make the transfer again and select all the required values.'))
            return None

        self.set_user_info_into_session(request, user_info)

        if not consumer.can_login():
            # The consumer does not login.
            return None

        if request.user.is_authenticated:
            openid2_id = user_info.get('openid2_id')
            if openid2_id is not None:
                if OIDCIdentifier.objects.filter(uident=openid2_id).exclude(user=request.user).exists():
                    messages.error(
                        request,
                        _("Pairing cannot be performed. This identifier is already paired with another account."))
                    return request.user
            self.create_identifier_if_missing(request, request.user, consumer.provider, openid2_id)
            return request.user
        return self.not_authenticated_user(request, user_info, consumer)

    def create_identifier_if_missing(self, request, user, provider, openid2_id):
        if openid2_id is not None:
            dummy, created = OIDCIdentifier.objects.get_or_create(
                uident=openid2_id, defaults=dict(user=user, provider=provider))
            if created:
                messages.success(request, _('The account has been successfully paired with the provider.'))

    def not_authenticated_user(self, request, user_info, consumer):
        # email based filtering
        users = self.filter_users_by_claims(user_info)
        openid2_id = user_info.get('openid2_id')
        if len(users) == 0:
            if openid2_id:
                users = [ident.user for ident in OIDCIdentifier.objects.filter(uident=openid2_id)]
        if len(users) == 1:
            user = users[0]
            self.create_identifier_if_missing(request, user, consumer.provider, openid2_id)
            if openid2_id:
                user = OIDCIdentifier.objects.get(uident=openid2_id).user
            if not user.is_active:
                msg = _("Your account is deactivated. Please contact our support.")
                messages.error(request, msg)
                return None
            return user
        elif len(users) > 1:
            # In the rare case that two user accounts have the same email address,
            # bail. Randomly selecting one seems really wrong.
            msg = 'Multiple users returned'
            raise SuspiciousOperation(msg)
        elif consumer.can_login():
            if consumer.no_new_user:
                msg = _("To pair with your identity provider, log in first.")
                messages.info(request, msg)
                return None
            return self.create_new_user(request, consumer, openid2_id, user_info)
        else:
            LOGGER.debug('Login failed: No user with email %s found.', user_info.get('email'))
            return None

    def create_new_user(self, request, consumer, openid2_id, user_info):
        """Create new user."""
        user = self.create_user(user_info)
        if openid2_id:
            OIDCIdentifier.objects.get_or_create(user=user, provider=consumer.provider, uident=openid2_id)
        msg = _("A new account has been created with the username {} and email {}.").format(
            user.username, user.email)
        messages.success(request, msg)
        return user



from django.test import SimpleTestCase

from djangocms_oidc.widgets import JsonDataTextarea


class JsonDataTextareaTest(SimpleTestCase):

    def setUp(self):
        self.widget = JsonDataTextarea()

    def test_none(self):
        self.assertIsNone(self.widget.format_value(None))

    def test_empty_string(self):
        self.assertIsNone(self.widget.format_value(""))

    def test_valid_json(self):
        response = self.widget.format_value('{"data": 42}')
        self.assertJSONEqual(response, {"data": 42})

    def test_invalid_json(self):
        data = '{data": 42}'
        self.assertEqual(self.widget.format_value(data), data)
        data = '{"data": 42'
        self.assertEqual(self.widget.format_value(data), data)
        data = '{"data": 42{'
        self.assertEqual(self.widget.format_value(data), data)
        data = '{"data": 42}}'
        self.assertEqual(self.widget.format_value(data), data)
        data = '{"data" 42}'
        self.assertEqual(self.widget.format_value(data), data)
        data = '{"data": 42} {"end": "test"}'
        self.assertEqual(self.widget.format_value(data), data)



from django.contrib import admin

from .forms import OIDCProviderForm


class OIDCIdentifierAdmin(admin.ModelAdmin):
    list_display = ('user', 'user_email', 'provider', 'uident')

    def user_email(self, obj):
        return obj.user.email


class OIDCRegisterConsumerAdmin(admin.ModelAdmin):
    list_display = ('name', 'register_url')


class OIDCProviderAdmin(admin.ModelAdmin):
    form = OIDCProviderForm



from cms.plugin_base import CMSPluginBase
from cms.plugin_pool import plugin_pool
from django.utils.translation import gettext_lazy as _

from .constants import DJANGOCMS_USER_SESSION_KEY
from .forms import OIDCDataForm, OIDCHandoverDataForm
from .helpers import check_required_handovered, get_user_identifiers_formset, get_verified_as
from .models import (
    OIDCDisplayDedicatedContent,
    OIDCHandoverData,
    OIDCIdentifier,
    OIDCLogin,
    OIDCShowAttribute,
    get_display_content_settings,
)


class OIDCConsumerBase(CMSPluginBase):

    module = _('OpenID Connect')
    form = OIDCDataForm
    change_form_template = "djangocms_oidc/change_form/consumer.html"
    text_enabled = True
    cache = False

    def get_verified_as(self, instance, user_info):
        default = _("User")
        if instance.verified_by is None:
            return user_info.get('email', default)
        return get_verified_as(instance.verified_by, user_info, default)

    def render(self, context, instance, placeholder):
        context = super().render(context, instance, placeholder)
        user_info = context['request'].session.get(DJANGOCMS_USER_SESSION_KEY)
        if user_info is not None:
            context['djangocms_oidc_user_info'] = user_info
            context['all_required_handovered'] = check_required_handovered(instance, user_info)
            context['djangocms_oidc_verified_as'] = self.get_verified_as(instance, user_info)
        context['registration_consumer_info'] = instance.provider.get_registration_consumer_info()
        return context


@plugin_pool.register_plugin
class OIDCHandoverDataPlugin(OIDCConsumerBase):
    name = _('OIDC Handover data')
    model = OIDCHandoverData
    form = OIDCHandoverDataForm
    render_template = "djangocms_oidc/handover_data.html"


@plugin_pool.register_plugin
class OIDCLoginPlugin(OIDCConsumerBase):
    name = _('OIDC Login')
    model = OIDCLogin
    render_template = "djangocms_oidc/login.html"


@plugin_pool.register_plugin
class OIDCListIdentifiersPlugin(CMSPluginBase):
    module = _('OpenID Connect')
    name = _('OIDC List identifiers')
    render_template = "djangocms_oidc/list_identifiers.html"
    change_form_template = "djangocms_oidc/change_form/list_identifiers.html"

    def render(self, context, instance, placeholder):
        context = super().render(context, instance, placeholder)
        if context['request'].user.is_authenticated:
            context['formset'] = get_user_identifiers_formset(context['request'].user)
            context['user_has_identifiers'] = OIDCIdentifier.objects.filter(user=context['request'].user).exists()
        return context


@plugin_pool.register_plugin
class OIDCDisplayDedicatedContentPlugin(CMSPluginBase):
    module = _('OpenID Connect')
    name = _('OIDC Display dedicated content')
    model = OIDCDisplayDedicatedContent
    render_template = "djangocms_oidc/display_dedicated_content.html"
    change_form_template = 'djangocms_oidc/change_form/display_dedicated_content.html'
    cache = False
    allow_children = True

    def render(self, context, instance, placeholder):
        context = super().render(context, instance, placeholder)
        user_info = context['request'].session.get(DJANGOCMS_USER_SESSION_KEY)
        if instance.conditions is None:
            context['content_permitted_to_user'] = user_info is not None
        else:
            for code, dummy, fnc in get_display_content_settings():
                if code == instance.conditions:
                    context['content_permitted_to_user'] = fnc(context, instance, placeholder, user_info)
        return context


@plugin_pool.register_plugin
class OIDCShowAttributePlugin(CMSPluginBase):
    module = _('OpenID Connect')
    name = _('OIDC Show attribute')
    model = OIDCShowAttribute
    render_template = "djangocms_oidc/show_attribute.html"
    cache = False

    def render(self, context, instance, placeholder):
        context = super().render(context, instance, placeholder)
        user_info = context['request'].session.get(DJANGOCMS_USER_SESSION_KEY)
        if user_info is not None:
            context['djangocms_oidc_verified_as'] = get_verified_as(
                instance.verified_by, user_info, instance.default_value)
        else:
            context['djangocms_oidc_verified_as'] = instance.default_value
        return context


@plugin_pool.register_plugin
class OIDCShowAttributeCountryPlugin(OIDCShowAttributePlugin):
    name = _('OIDC Show attribute Country')
    render_template = "djangocms_oidc/show_attribute_country.html"



from distutils.command.build import build

from setuptools import find_packages, setup


class CustomBuild(build):
    sub_commands = [("compile_catalog", lambda x: True)] + build.sub_commands


setup(
    author='Zdeněk Böhm',
    author_email='zdenek.bohm@nic.cz',
    name='djangocms-oidc',
    version='4.1.0',
    description='Plugin OIDC (OpenID Connect) into Django CMS.',
    long_description=open('README.rst').read(),
    long_description_content_type='text/x-rst',
    url='https://github.com/CZ-NIC/djangocms-oidc',
    license='GPL GNU License',
    platforms=['OS Independent'],
    classifiers=(
        'Development Status :: 5 - Production/Stable',
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU General Public License (GPL)',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
        'Topic :: Software Development',
        'Topic :: Software Development :: Libraries :: Application Frameworks',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Framework :: Django CMS :: 3.11',
    ),
    install_requires=(
        'django-cms~=3.11',
        'mozilla-django-oidc~=3.0',
        'django-countries~=7.5',
        'django-multiselectfield',
    ),
    extras_require={
        'quality': ['isort', 'flake8'],
        'test': ['requests_mock', 'freezegun'],
    },
    packages=find_packages(exclude=['tests']),
    cmdclass={"build": CustomBuild},
    setup_requires=["Babel >=2.3"],
    include_package_data=True,
    zip_safe=False
)



from django.contrib.admin.sites import site
from django.contrib.auth.models import AnonymousUser
from django.test import RequestFactory, TestCase

from djangocms_oidc.admin.options import OIDCProviderAdmin
from djangocms_oidc.models import OIDCProvider, OIDCRegisterConsumer


class TestAdminForms(TestCase):

    post = {
        'name': 'Name',
        'slug': 'name',
        'authorization_endpoint': 'https://foo.foo/auth',
        'token_endpoint': 'https://foo.foo/token',
        'user_endpoint': 'https://foo.foo/user',
    }
    error_message = {
        'client_id': ['The value is required if "Register consumer" is not set.'],
        'client_secret': ['The value is required if "Register consumer" is not set.'],
    }

    def get_form(self):
        provider_admin = OIDCProviderAdmin(admin_site=site, model=OIDCProvider)
        request = RequestFactory().request()
        request.user = AnonymousUser()
        return provider_admin.get_form(request)

    def test_form(self):
        post = self.post.copy()
        post.update({
            'client_id': '42',
            'client_secret': 'secret',
        })
        OIDCProviderForm = self.get_form()
        form = OIDCProviderForm(post)
        self.assertTrue(form.is_valid())

    def test_form_client_id_missing(self):
        post = self.post.copy()
        post['client_secret'] = 'secret'
        OIDCProviderForm = self.get_form()
        form = OIDCProviderForm(post)
        self.assertEqual(form.errors, self.error_message)

    def test_form_client_secret_missing(self):
        post = self.post.copy()
        post['client_id'] = '42'
        OIDCProviderForm = self.get_form()
        form = OIDCProviderForm(post)
        self.assertEqual(form.errors, self.error_message)

    def test_form_with_register_consumer(self):
        post = self.post.copy()
        post['register_consumer'] = OIDCRegisterConsumer.objects.create(register_url="https://foo.foo/reg")
        OIDCProviderForm = self.get_form()
        form = OIDCProviderForm(post)
        self.assertTrue(form.is_valid())



import datetime

import requests_mock
from django.core import cache
from django.core.exceptions import ValidationError
from django.test import SimpleTestCase, TestCase, override_settings
from django.utils import timezone
from freezegun import freeze_time

from djangocms_oidc.models import (
    ConsumerRegistrationExpired,
    OIDCHandoverData,
    OIDCProvider,
    OIDCRegisterConsumer,
    OIDCShowAttribute,
    validate_claims,
)


class TestOIDCRegisterConsumer(SimpleTestCase):

    def test_str(self):
        consumer = OIDCRegisterConsumer(name="Test")
        self.assertEqual(str(consumer), 'Test')

    def test_get_payload(self):
        consumer = OIDCRegisterConsumer(name="Test")
        payload = consumer.get_payload('The Name', ['https://foo.foo/1', 'https://foo.foo/2'])
        self.assertEqual(payload, {
            'application_type': 'web',
            'id_token_signed_response_alg': 'HS256',
            'redirect_uris': ['https://foo.foo/1', 'https://foo.foo/2'],
            'token_endpoint_auth_method': 'client_secret_post',
            'response_types': ['code'],
            'client_name': 'The Name'
        })

    @requests_mock.Mocker()
    def test_make_registration(self, mock_req):
        mock_req.post("https://foo.foo/register", json={'status': 'OK'})
        consumer = OIDCRegisterConsumer(name="Test", register_url="https://foo.foo/register")
        response = consumer.make_registration('The Name', ['https://foo.foo/1', 'https://foo.foo/2'])
        self.assertEqual(response, {'status': 'OK'})


@override_settings(CMS_CACHE_PREFIX='prefix:')
@override_settings(CACHES={'default': {'BACKEND': 'django.core.cache.backends.locmem.LocMemCache'}})
class TestOIDCProvider(TestCase):

    @classmethod
    def setUpTestData(cls):
        cls.register_consumer = OIDCRegisterConsumer.objects.create(
            name="Test", register_url="https://foo.foo/register")

    def setUp(self):
        cache.cache.clear()

    def test_str(self):
        provider = OIDCProvider(name="Provider")
        self.assertEqual(str(provider), 'Provider')

    def test_get_client_id(self):
        provider = OIDCProvider(name="Provider", client_id="42")
        self.assertEqual(provider.get_client_id(), "42")

    def test_get_client_id_cache_expired(self):
        provider = OIDCProvider(name="Provider")
        with self.assertRaises(ConsumerRegistrationExpired):
            provider.get_client_id()

    def test_get_client_id_from_cache(self):
        provider = OIDCProvider.objects.create(name="Provider", slug="provider")
        cache.cache.set(f'prefix:djangocms_oidc_provider:{provider.pk}', {'client_id': 42})
        client_id = provider.get_client_id()
        self.assertEqual(client_id, 42)

    def test_get_client_secret(self):
        provider = OIDCProvider(name="Provider", client_secret="42")
        self.assertEqual(provider.get_client_secret(), "42")

    def test_get_client_secret_cache_expired(self):
        provider = OIDCProvider(name="Provider")
        with self.assertRaises(ConsumerRegistrationExpired):
            provider.get_client_secret()

    def test_get_client_secret_from_cache(self):
        provider = OIDCProvider.objects.create(name="Provider", slug="provider")
        cache.cache.set(f'prefix:djangocms_oidc_provider:{provider.pk}', {'client_secret': 42})
        get_client_secret = provider.get_client_secret()
        self.assertEqual(get_client_secret, 42)

    def test_needs_register(self):
        provider = OIDCProvider(name="Provider", client_id=42)
        self.assertFalse(provider.needs_register())

    def test_needs_register_from_cache(self):
        provider = OIDCProvider.objects.create(name="Provider", slug="provider")
        cache.cache.set(f'prefix:djangocms_oidc_provider:{provider.pk}', {'client_id': 42})
        self.assertFalse(provider.needs_register())

    def test_needs_register_is_true(self):
        provider = OIDCProvider(name="Provider")
        self.assertTrue(provider.needs_register())

    def test_get_cache_key(self):
        provider = OIDCProvider.objects.create(name="Provider", slug="provider")
        self.assertEqual(provider.get_cache_key(), f'prefix:djangocms_oidc_provider:{provider.pk}')

    def test_registration_in_progress_not(self):
        provider = OIDCProvider(name="Provider", client_id=42)
        self.assertFalse(provider.registration_in_progress())

    def test_registration_in_progress(self):
        provider = OIDCProvider.objects.create(name="Provider", slug="provider")
        cache.cache.set(f'prefix:djangocms_oidc_provider:{provider.pk}', '--PROGRESS--')
        self.assertTrue(provider.registration_in_progress())

    @requests_mock.Mocker()
    def test_register_consumer_into_cache_request_exception(self, mock_req):
        mock_req.post("https://foo.foo/register", text='Not Found', status_code=404)
        provider = OIDCProvider.objects.create(
            name="Provider", slug="provider", register_consumer=self.register_consumer)
        with freeze_time('2020-10-06 14:55') as frozen_datetime:
            msg = provider.register_consumer_into_cache(['https://foo.foo/1/', 'https://foo.foo/2/'])
            self.assertEqual(cache.cache.get(f'prefix:djangocms_oidc_provider:{provider.pk}'), '--PROGRESS--')
            frozen_datetime.tick(delta=datetime.timedelta(seconds=10))
            self.assertIsNone(cache.cache.get(f'prefix:djangocms_oidc_provider:{provider.pk}'))
        self.assertEqual(str(msg), '404 Client Error: None for url: https://foo.foo/register')

    @requests_mock.Mocker()
    def test_register_consumer_into_cache_expire_never(self, mock_req):
        mock_req.post("https://foo.foo/register", json={
            'client_secret_expires_at': 0,
            'client_secret': '5ed3a93b14cb5f1fdb97cb3dde1daddade443753eeb84432320b78a2',
            'client_id': '1d9G0Oid4E5V'
        })
        provider = OIDCProvider.objects.create(
            name="Provider", slug="provider", register_consumer=self.register_consumer)
        msg = provider.register_consumer_into_cache(['https://foo.foo/1/', 'https://foo.foo/2/'])
        self.assertIsNone(msg)
        data = cache.cache.get(f'prefix:djangocms_oidc_provider:{provider.pk}')
        self.assertEqual(data, {
            'client_id': '1d9G0Oid4E5V',
            'client_secret': '5ed3a93b14cb5f1fdb97cb3dde1daddade443753eeb84432320b78a2',
            'expires_at': 'never',
        })

    @freeze_time('2020-10-06 15:00:00')
    @requests_mock.Mocker()
    def test_register_consumer_into_cache(self, mock_req):
        mock_req.post("https://foo.foo/register", json={
            'client_id_issued_at': 1601989200,
            'response_types': ['code'],
            'request_object_signing_alg': 'HS256',
            'registration_client_uri': 'https://mojeid.regtest.nic.cz/oidc/registration/?client_id=1d9G0Oid4E5V',
            'application_type': 'web',
            'registration_access_token': 'zhbvda8JZgsamO8Lm0TgZ0UA50cLXrZm',
            'redirect_uris': ['http://localhost/oidc/callback/'],
            'client_secret_expires_at': 1602075600,
            'client_secret': '5ed3a93b14cb5f1fdb97cb3dde1daddade443753eeb84432320b78a2',
            'client_name': 'MojeID Test - Dynamic consumer',
            'client_id': '1d9G0Oid4E5V'
        })
        provider = OIDCProvider.objects.create(
            name="Provider", slug="provider", register_consumer=self.register_consumer)
        msg = provider.register_consumer_into_cache(['https://foo.foo/1/', 'https://foo.foo/2/'])
        self.assertIsNone(msg)
        data = cache.cache.get(f'prefix:djangocms_oidc_provider:{provider.pk}')
        self.assertEqual(data, {
            'client_id': '1d9G0Oid4E5V',
            'client_secret': '5ed3a93b14cb5f1fdb97cb3dde1daddade443753eeb84432320b78a2',
            'expires_at': datetime.datetime(2020, 10, 7, 13, 0, tzinfo=timezone.utc),
        })

    def test_get_registration_consumer_info_managed(self):
        provider = OIDCProvider.objects.create(name="Provider", slug="provider", client_id=42)
        data = provider.get_registration_consumer_info()
        self.assertEqual(data, {'client_id': 42, 'expires_at': None, 'consumer_type': 'MANAGED'})

    def test_get_registration_consumer_info_automatic(self):
        provider = OIDCProvider.objects.create(name="Provider", slug="provider")
        data = provider.get_registration_consumer_info()
        self.assertEqual(data, {'client_id': None, 'expires_at': None, 'consumer_type': 'AUTOMATIC'})

    def test_get_registration_consumer_info_automatic_from_cache(self):
        provider = OIDCProvider.objects.create(name="Provider", slug="provider")
        cache.cache.set(f'prefix:djangocms_oidc_provider:{provider.pk}', {
            'client_id': 42, 'expires_at': 'never'})
        data = provider.get_registration_consumer_info()
        self.assertEqual(data, {'client_id': 42, 'expires_at': 'never', 'consumer_type': 'AUTOMATIC'})


class TestValidateClaims(SimpleTestCase):

    def test_no_userinfo(self):
        validate_claims({})

    def test_no_dict_type(self):
        with self.assertRaisesMessage(ValidationError, "The value must be a dictionary type: {}"):
            validate_claims('')

    def test_userinfo_is_str(self):
        with self.assertRaisesMessage(
                ValidationError, 'The value "userinfo" must be a dictionary type: {"userinfo": ...}'):
            validate_claims({'userinfo': 'foo'})

    def test_userinfo_item_is_str(self):
        with self.assertRaisesMessage(
                ValidationError, 'The "name" must be a dictionary type: "name": {"essential": true},'):
            validate_claims({'userinfo': {'name': 'foo'}})

    def test_userinfo_essential_missing(self):
        with self.assertRaisesMessage(
                ValidationError, 'The "name" value must contain key "essential": "name": {"essential": true},'):
            validate_claims({'userinfo': {'name': {'value': 'foo'}}})

    def test_userinfo_essential_is_not_boolean(self):
        with self.assertRaisesMessage(
                ValidationError, 'The value of "essential" must be a boolean type": "name": {"essential": true},'):
            validate_claims({'userinfo': {'name': {'essential': 'foo'}}})

    def test_userinfo_with_claims(self):
        validate_claims({'userinfo': {'name': {'essential': True}}})


class TestOIDCHandoverData(SimpleTestCase):

    def test_provider(self):
        provider = OIDCProvider(name="Provider")
        consumer = OIDCHandoverData(provider=provider)
        self.assertEqual(str(consumer), "Provider")

    def test_no_provider(self):
        consumer = OIDCHandoverData()
        self.assertEqual(str(consumer), "[No provider yet]")

    def test_button(self):
        consumer = OIDCHandoverData(button_label="Foo")
        self.assertEqual(str(consumer), "Foo")


class TestOIDCShowAttribute(SimpleTestCase):

    def test(self):
        show_arrib = OIDCShowAttribute(verified_by="Provider")
        self.assertEqual(str(show_arrib), "Provider")



import json
import logging
from urllib.parse import urlencode, urlparse

from django.contrib import messages
from django.contrib.auth import logout
from django.http import HttpResponseRedirect, JsonResponse
from django.urls import reverse
from django.utils.crypto import get_random_string
from django.utils.translation import gettext_lazy as _
from django.views.generic import View
from mozilla_django_oidc.utils import absolutify, add_state_and_nonce_to_session, import_from_settings
from mozilla_django_oidc.views import OIDCAuthenticationCallbackView, OIDCAuthenticationRequestView, get_next_url

from .constants import DJANGOCMS_PLUGIN_SESSION_KEY, DJANGOCMS_USER_SESSION_KEY
from .helpers import get_consumer, get_user_identifiers_formset, load_consumer, request_is_ajax, set_consumer

LOGGER = logging.getLogger("djangocms_oidc")


class OIDCSignupView(View):
    """View for all consumer types - Login or Handover data."""

    http_method_names = ['get']

    @staticmethod
    def get_settings(attr, *args):
        return import_from_settings(attr, *args)

    def get_redirect_params(self, request, prompt):
        parsed = urlparse(request.META.get('HTTP_REFERER', '/'))
        params = {
            self.get_settings('OIDC_REDIRECT_FIELD_NAME', 'next'): parsed.path
        }
        if prompt:
            params['prompt'] = prompt
        return params

    def get(self, request, consumer_type, plugin_id, prompt=None):
        """Keep plugin ID and redirect to the OIDC consumer."""
        consumer = load_consumer(consumer_type, plugin_id)
        if consumer is None:
            messages.error(request, _('OIDC Consumer activation falied. Please try again.'))
            return HttpResponseRedirect(request.META.get('HTTP_REFERER', '/'))
        if consumer.provider.registration_in_progress():
            messages.info(request, _('Communication with the provider is in progress. Please try again later.'))
            return HttpResponseRedirect(request.META.get('HTTP_REFERER', '/'))
        if consumer.provider.needs_register():
            redirect_uris = [request.build_absolute_uri(reverse("oidc_authentication_callback"))]
            error = consumer.provider.register_consumer_into_cache(redirect_uris)
            if error:
                messages.error(request, _('Communication with the provider failed. Please try again later.'))
                messages.error(request, error)
                return HttpResponseRedirect(request.META.get('HTTP_REFERER', '/'))
        set_consumer(request, consumer)
        if DJANGOCMS_USER_SESSION_KEY in request.session:
            del request.session[DJANGOCMS_USER_SESSION_KEY]
        query = urlencode(self.get_redirect_params(request, prompt))
        redirect_url = '{url}?{query}'.format(url=reverse("oidc_authentication_init"), query=query)
        return HttpResponseRedirect(redirect_url)


class OIDCDismissView(View):
    """Dismiss handovered data."""

    http_method_names = ['get']

    def get(self, request):
        """Remove OIDC data."""
        if DJANGOCMS_PLUGIN_SESSION_KEY in request.session:
            del request.session[DJANGOCMS_PLUGIN_SESSION_KEY]
        if DJANGOCMS_USER_SESSION_KEY in request.session:
            del request.session[DJANGOCMS_USER_SESSION_KEY]
        return HttpResponseRedirect(request.META.get('HTTP_REFERER', '/'))


class OIDCLogoutView(View):
    """Logout view. Return back to the current page after logim. Not to the page with Login form as LoginView does."""

    http_method_names = ['get']

    def get(self, request):
        """Logout user. As a result will be lost also handovered data."""
        if request.user.is_authenticated:
            logout(request)
        return HttpResponseRedirect(request.META.get('HTTP_REFERER', '/'))


class OIDCDeleteIdentifiersView(View):
    """Delete identifier connected the user with the provider."""

    http_method_names = ['post']

    def post(self, request, *args, **kwargs):
        if request.user.is_authenticated:
            data = {'result': "NOCHANGE"}
            formset = get_user_identifiers_formset(request.user, request.POST, request.FILES)
            if formset.is_valid():
                formset.save()
                if formset.deleted_objects:
                    data['result'] = "SUCCESS"
                    data['messages'] = [_("Identifier has been deleted.")]
            else:
                data['result'] = "ERROR"
                data['messages'] = formset.errors
            if request_is_ajax(request):
                return JsonResponse(data)
            elif data['result'] == "SUCCESS":
                messages.success(request, _("Identifier has been deleted."))
            elif data['result'] == "NOCHANGE":
                messages.info(request, _("No identifier has been deleted. Check one to delete."))
            else:
                for msg in data['messages']:
                    messages.error(request, msg.as_text())
        return HttpResponseRedirect(request.META.get('HTTP_REFERER', '/'))


class DjangocmsOIDCAuthenticationRequestView(OIDCAuthenticationRequestView):

    def __init__(self, *args, **kwargs):
        # The function is necessary due to supress required OIDC_OP_AUTHORIZATION_ENDPOINT and OIDC_RP_CLIENT_ID
        # in OIDCAuthenticationRequestView.
        super(View, self).__init__(*args, **kwargs)

    def get_extra_params(self, request, consumer):
        params = {}
        if consumer is None:
            claims = self.get_settings('OIDC_AUTH_REQUEST_EXTRA_PARAMS', {})
            if claims:
                params.update(claims)
        else:
            params["claims"] = json.dumps(consumer.claims)
            prompt = set()
            if consumer.authorization_prompt:
                prompt = set(consumer.authorization_prompt)
            prompt_get = request.GET.get('prompt')
            if prompt_get:
                choices = [code for code, _ in consumer.AUTHORIZATION_PROMPT]
                for code in prompt_get.split(","):
                    if code in choices:
                        prompt.add(code)
            if prompt:
                if "none" in prompt and len(prompt) > 1:
                    prompt.remove("none")
                params['prompt'] = " ".join(prompt)
        return params

    def get(self, request):
        """OIDC client authentication initialization HTTP endpoint"""
        state = get_random_string(self.get_settings('OIDC_STATE_SIZE', 32))
        redirect_field_name = self.get_settings('OIDC_REDIRECT_FIELD_NAME', 'next')
        reverse_url = self.get_settings('OIDC_AUTHENTICATION_CALLBACK_URL',
                                        'oidc_authentication_callback')

        consumer = get_consumer(request)
        if consumer is None:
            messages.error(request, _('OIDC Consumer activation falied. Please try again.'))
            return HttpResponseRedirect(request.META.get('HTTP_REFERER', '/'))
        params = {
            'response_type': 'code',
            'scope': self.get_settings('OIDC_RP_SCOPES', 'openid email'),
            'client_id': consumer.provider.get_client_id(),  # OIDC_RP_CLIENT_ID
            'redirect_uri': absolutify(
                request,
                reverse(reverse_url)
            ),
            'state': state,
        }
        if self.get_settings('OIDC_USE_NONCE', True):
            nonce = get_random_string(self.get_settings('OIDC_NONCE_SIZE', 32))
            params.update({
                'nonce': nonce
            })
            request.session['oidc_nonce'] = nonce

        add_state_and_nonce_to_session(request, state, params)

        if consumer.redirect_page is None:
            request.session['oidc_login_next'] = get_next_url(request, redirect_field_name)
        else:
            request.session['oidc_login_next'] = consumer.redirect_page.get_absolute_url()

        params.update(self.get_extra_params(request, consumer))
        query = urlencode(params)
        redirect_url = f'{consumer.provider.authorization_endpoint}?{query}'
        return HttpResponseRedirect(redirect_url)


class DjangocmsOIDCAuthenticationCallbackView(OIDCAuthenticationCallbackView):

    @property
    def failure_url(self):
        consumer = get_consumer(self.request)
        if consumer is not None and consumer.can_login():
            messages.info(self.request, _("Login failed."))
        next_url = self.request.session.get('oidc_login_next', None)
        return self.get_settings('LOGIN_REDIRECT_URL_FAILURE', '/') if next_url is None else next_url



from django.http import HttpResponse
from django.views.generic import View


class TestHomePage(View):

    def get(self, request, *args, **kwargs):
        return HttpResponse(b"Test Home page.")


class TestLoginPage(View):

    def get(self, request, *args, **kwargs):
        return HttpResponse(b"Test Login page.")


class TestFakePage(View):

    def get(self, request, *args, **kwargs):
        return HttpResponse(b"Win!")
